import pytest
import shutil
import os
import glob
import subprocess
import time as ttime
import multiprocessing

from bluesky_queueserver.manager.profile_ops import gen_list_of_plans_and_devices
from bluesky_queueserver.manager.comms import zmq_single_request

from .common import (
    copy_default_profile_collection,
    append_code_to_last_startup_file,
    wait_for_condition,
    condition_environment_created,
    condition_queue_processing_finished,
    condition_environment_closed,
)

from .common import re_manager_cmd  # noqa: F401

# User name and user group name used throughout most of the tests.
_user, _user_group = "Testing Script", "admin"

_plan1 = {"name": "count", "args": [["det1", "det2"]], "item_type": "plan"}

_sample_plan1 = """
def simple_sample_plan():
    '''
    Simple plan for tests.
    '''
    yield from count([det1, det2])
"""


# fmt: off
@pytest.mark.parametrize("option", ["startup_dir", "profile", "multiple"])
# fmt: on
def test_manager_options_startup_profile(re_manager_cmd, tmp_path, monkeypatch, option):  # noqa: F811
    pc_path = copy_default_profile_collection(tmp_path)

    # Add extra plan. The original set of startup files will not contain this plan.
    append_code_to_last_startup_file(pc_path, additional_code=_sample_plan1)

    # Generate the new list of allowed plans and devices and reload them
    gen_list_of_plans_and_devices(startup_dir=pc_path, file_dir=pc_path, overwrite=True)

    # Start manager
    if option == "startup_dir":
        re_manager_cmd(["--startup-dir", pc_path])
    elif option == "profile":
        # This option is more complicated: we want to recreate the structure of IPython startup
        #   directory: <some root dir>/profile_<profile_name>/startup.
        root_dir = os.path.split(pc_path)[0]
        monkeypatch.setenv("IPYTHONDIR", root_dir)
        profile_name = "testing"
        startup_path = os.path.join(root_dir, f"profile_{profile_name}", "startup")
        os.makedirs(startup_path)

        file_pattern = os.path.join(pc_path, "*")
        for fl_path in glob.glob(file_pattern):
            shutil.move(fl_path, startup_path)

        os.rmdir(pc_path)

        # We pass only profile name as a parameter.
        re_manager_cmd(["--startup-profile", profile_name])
    elif option == "multiple":
        # Expected to fail if multiple options are selected.
        with pytest.raises(TimeoutError, match="RE Manager failed to start"):
            re_manager_cmd(["--startup-dir", pc_path, "--startup-profile", "some_name"])
        return
    else:
        assert False, f"Unknown option '{option}'"

    # Open the environment (make sure that the environment loads)
    resp1, _ = zmq_single_request("environment_open")
    assert resp1["success"] is True
    assert wait_for_condition(time=10, condition=condition_environment_created)

    # Add the plan to the queue (will fail if incorrect environment is loaded)
    plan = {"name": "simple_sample_plan", "item_type": "plan"}
    params = {"item": plan, "user": _user, "user_group": _user_group}
    resp2, _ = zmq_single_request("queue_item_add", params)
    assert resp2["success"] is True, f"resp={resp2}"

    # Start the queue
    resp3, _ = zmq_single_request("queue_start")
    assert resp3["success"] is True
    assert wait_for_condition(time=10, condition=condition_queue_processing_finished)

    # Make sure that the plan was executed
    resp4, _ = zmq_single_request("status")
    assert resp4["items_in_queue"] == 0
    assert resp4["items_in_history"] == 1

    # Close the environment
    resp5, _ = zmq_single_request("environment_close")
    assert resp5["success"] is True, f"resp={resp5}"
    assert wait_for_condition(time=5, condition=condition_environment_closed)

    monkeypatch.setenv("IPYTHONDIR", "abc")


@pytest.fixture
def zmq_proxy():
    cmd = ["bluesky-0MQ-proxy", "5567", "5568"]
    p = subprocess.Popen(cmd, universal_newlines=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    yield
    p.kill()


class _StartDispatcherProcess(multiprocessing.Process):
    def __init__(self, queue, **kwargs):
        super().__init__(**kwargs)
        self._queue = queue

    def run(self):
        def put_in_queue(name, doc):
            print("putting ", name, "in queue")
            self._queue.put((name, doc))

        from bluesky.callbacks.zmq import RemoteDispatcher

        d = RemoteDispatcher("127.0.0.1:5568")
        d.subscribe(put_in_queue)
        print("REMOTE IS READY TO START")
        d.loop.call_later(9, d.stop)
        d.start()


@pytest.fixture
def zmq_dispatcher():
    # The following code is refactored code from 'bluesky.tests.test_zmq.py' (test_zmq_no_RE)
    queue = multiprocessing.Queue()
    dispatcher_proc = _StartDispatcherProcess(queue=queue, daemon=True)
    dispatcher_proc.start()
    ttime.sleep(2)  # As above, give this plenty of time to start.

    yield queue

    dispatcher_proc.terminate()
    dispatcher_proc.join()


def test_manager_acq_with_0MQ_proxy(re_manager_cmd, zmq_proxy, zmq_dispatcher):  # noqa: F811
    re_manager_cmd(["--zmq-data-proxy-addr", "localhost:5567"])

    # Open the environment (make sure that the environment loads)
    resp1, _ = zmq_single_request("environment_open")
    assert resp1["success"] is True
    assert wait_for_condition(time=10, condition=condition_environment_created)

    # Add the plan to the queue (will fail if incorrect environment is loaded)
    params = {"item": _plan1, "user": _user, "user_group": _user_group}
    resp2, _ = zmq_single_request("queue_item_add", params)
    assert resp2["success"] is True, f"resp={resp2}"

    # Start the queue
    resp3, _ = zmq_single_request("queue_start")
    assert resp3["success"] is True
    assert wait_for_condition(time=10, condition=condition_queue_processing_finished)

    # Make sure that the plan was executed
    resp4, _ = zmq_single_request("status")
    assert resp4["items_in_queue"] == 0
    assert resp4["items_in_history"] == 1

    # Close the environment
    resp5, _ = zmq_single_request("environment_close")
    assert resp5["success"] is True, f"resp={resp5}"
    assert wait_for_condition(time=5, condition=condition_environment_closed)

    # Test if the data was delivered to the consumer.
    # Simple test: check if 'start' and 'stop' documents were delivered.
    queue = zmq_dispatcher
    remote_accumulator = []
    while not queue.empty():  # Since queue is used by one process at a time, queue.empty() should work reliably
        remote_accumulator.append(queue.get(timeout=2))
    assert len(remote_accumulator) >= 2
    assert remote_accumulator[0][0] == "start"  # Start document
    assert remote_accumulator[-1][0] == "stop"  # Stop document


# fmt: off
@pytest.mark.parametrize("redis_addr, success", [
    ("localhost", True),
    ("localhost:6379", True),
    ("localhost:6378", False)])  # Incorrect port.
# fmt: on
def test_manager_redis_addr_parameter(re_manager_cmd, redis_addr, success):  # noqa: F811
    if success:
        re_manager_cmd(["--redis-addr", redis_addr])

        # Try to communicate with the server to make sure Redis is configure correctly.
        #   RE Manager has to access Redis in order to prepare 'status'.
        resp1, _ = zmq_single_request("status")
        assert resp1["items_in_queue"] == 0
        assert resp1["items_in_history"] == 0
    else:
        with pytest.raises(TimeoutError, match="RE Manager failed to start"):
            re_manager_cmd(["--redis-addr", redis_addr])



import os
import pytest
import subprocess
import pprint

from ..comms import zmq_single_request
from .common import re_manager_cmd  # noqa: F401

from .common import (
    wait_for_condition,
    condition_environment_created,
    condition_environment_closed,
    condition_queue_processing_finished,
    copy_default_profile_collection,
    clear_redis_pool,
)

from bluesky_queueserver.manager.profile_ops import gen_list_of_plans_and_devices


# Plans used in most of the tests: '_plan1' and '_plan2' are quickly executed '_plan3' runs for 5 seconds.
_plan1 = {"name": "count", "args": [["det1", "det2"]], "item_type": "plan"}
_plan2 = {"name": "scan", "args": [["det1", "det2"], "motor", -1, 1, 10], "item_type": "plan"}
_plan3 = {"name": "count", "args": [["det1", "det2"]], "kwargs": {"num": 5, "delay": 1}, "item_type": "plan"}
_instruction_stop = {"name": "queue_stop", "item_type": "instruction"}

# User name and user group name used throughout most of the tests.
_user, _user_group = "Testing Script", "admin"

timeout_env_open = 10


# fmt: off
@pytest.mark.parametrize("option", ["--verbose", "--quiet", "--silent"])
# fmt: on
def test_start_re_manager_logging_1(re_manager_cmd, option):  # noqa: F811
    """
    Test if RE Manager is correctly started with parameters that define logging verbosity.
    The test also creates the worker environment to make sure that the program does not crash
    when worker process is created.

    This is a smoke test: it does not verify that logging works.
    """
    re_manager_cmd([option])

    resp1, _ = zmq_single_request("environment_open")
    assert resp1["success"] is True
    assert resp1["msg"] == ""

    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    # Attemtp to communicate with RE Manager
    resp2, _ = zmq_single_request("status")
    assert resp2["items_in_queue"] == 0
    assert resp2["items_in_history"] == 0

    resp3, _ = zmq_single_request("environment_close")
    assert resp3["success"] is True
    assert resp3["msg"] == ""

    assert wait_for_condition(time=3, condition=condition_environment_closed)


# fmt: off
@pytest.mark.parametrize("console_print, console_zmq", [
    (None, None),
    (True, True),
    (True, False),
    (False, True),
    (False, False),
])
# fmt: on
def test_start_re_manager_console_output_1(re_manager_cmd, console_print, console_zmq):  # noqa: F811
    """
    Test for printing and publishing the console output (--console-output and --zmq-publish).
    """
    params = []
    if console_print is True:
        params.extend(["--console-output", "ON"])
    elif console_print is False:
        params.extend(["--console-output", "OFF"])
    if console_zmq is True:
        params.extend(["--zmq-publish-console", "ON"])
    elif console_zmq is False:
        params.extend(["--zmq-publish-console", "OFF"])

    # Default values (if parameters are not specified)
    if console_print is None:
        console_print = True
    if console_zmq is None:
        console_zmq = False

    # Start monitor (captures messages published to 0MQ)
    p_monitor = subprocess.Popen(
        ["qserver-console-monitor"], universal_newlines=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )

    re_manager = re_manager_cmd(params, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    resp1, _ = zmq_single_request("environment_open")
    assert resp1["success"] is True
    assert resp1["msg"] == ""

    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    resp2, _ = zmq_single_request("queue_item_add", {"item": _plan1, "user": _user, "user_group": _user_group})
    assert resp2["success"] is True

    resp3, _ = zmq_single_request("queue_start")
    assert resp3["success"] is True

    assert wait_for_condition(time=10, condition=condition_queue_processing_finished)

    resp3, _ = zmq_single_request("environment_close")
    assert resp3["success"] is True
    assert resp3["msg"] == ""

    assert wait_for_condition(time=3, condition=condition_environment_closed)

    p = re_manager._p
    re_manager.stop_manager()
    re_manager_stdout, re_manager_stderr = p.communicate()

    p_monitor.terminate()
    streamed_stdout, streamed_stderr = p_monitor.communicate()

    def check_output_contents(collected_stdout):
        assert collected_stdout != ""
        # Verify that output from all sources is present in the output
        # Logging from manager
        assert "bluesky_queueserver.manager.manager manager:" in collected_stdout
        assert "bluesky_queueserver.manager.profile_ops profile_ops:" in collected_stdout
        # Logging from Worker
        assert "bluesky_queueserver.manager.worker worker:" in collected_stdout
        assert "RE Environment is ready" in collected_stdout
        # Printing from live table
        assert "generator count" in collected_stdout
        assert "Run was closed:" in collected_stdout

    assert re_manager_stderr == ""
    assert "bluesky_queueserver.manager.output_streaming" in streamed_stderr

    if console_print:
        check_output_contents(re_manager_stdout)
    else:
        assert re_manager_stdout == ""

    if console_zmq:
        check_output_contents(streamed_stdout)
    else:
        assert streamed_stdout == ""


# fmt: off
@pytest.mark.parametrize("option", ["unchanged", "add_plan", "add_device", "add_plan_device"])
@pytest.mark.parametrize("update_existing_plans_devices", ["NEVER", "ENVIRONMENT_OPEN", "ALWAYS"])
# fmt: on
def test_cli_update_existing_plans_devices_01(
    re_manager_cmd, tmp_path, update_existing_plans_devices, option  # noqa: F811
):
    """
    Testing the modes defined by ``--update-existing-plans-devices`` parameter: create a copy of
    profile collection, generate the list of existing plans and devices from startup files, add
    a plan and/or a device to startup files, start RE Manager, open and close the environment
    (it is expected to generate the updated lists of existing plans and devices and possibly
    save them to disk), verify that the new plan and/or device is in the lists of allowed plans
    and devices, reload permissions and existing plans and devices from disk and verify that
    the new plan and/or devices is in the list of allowed plans (in case the new list of existing
    devices is saved to file).
    """
    # Copy the default profile collection and generate the list of existing devices
    pc_path = copy_default_profile_collection(tmp_path, copy_yaml=True)
    gen_list_of_plans_and_devices(
        startup_dir=pc_path, file_dir=pc_path, file_name="existing_plans_and_devices.yaml", overwrite=True
    )

    # Start the manager
    params = ["--startup-dir", pc_path, "--update-existing-plans-devices", update_existing_plans_devices]
    re_manager_cmd(params)

    resp1, _ = zmq_single_request("status")
    devices_allowed_uid1 = resp1["devices_allowed_uid"]
    plans_allowed_uid1 = resp1["plans_allowed_uid"]

    # Add a plan ('count50') and a device ('det50') if needed
    with open(os.path.join(pc_path, "zz.py"), "w") as f:
        if option in ("add_device", "add_plan_device"):
            f.writelines("det50 = det\n")
        if option in ("add_plan", "add_plan_device"):
            f.writelines("count50 = count\n")

    resp2, _ = zmq_single_request("environment_open")
    assert resp2["success"] is True
    assert resp2["msg"] == ""

    assert wait_for_condition(time=timeout_env_open, condition=condition_environment_created)

    resp3, _ = zmq_single_request("environment_close")
    assert resp3["success"] is True
    assert resp3["msg"] == ""

    assert wait_for_condition(time=3, condition=condition_environment_closed)

    resp4, _ = zmq_single_request("status")
    devices_allowed_uid2 = resp4["devices_allowed_uid"]
    plans_allowed_uid2 = resp4["plans_allowed_uid"]

    def verify_allowed_lists(*, new_plan_added, new_device_added):
        resp5a, _ = zmq_single_request("plans_allowed", params={"user_group": _user_group})
        assert resp5a["success"] is True, f"resp={resp5a}"
        plans_allowed = resp5a["plans_allowed"]
        resp5b, _ = zmq_single_request("devices_allowed", params={"user_group": _user_group})
        assert resp5b["success"] is True, f"resp={resp5b}"
        devices_allowed = resp5b["devices_allowed"]

        if new_device_added:
            assert "det50" in devices_allowed
        else:
            assert "det50" not in devices_allowed
        if new_plan_added:
            assert "count50" in plans_allowed
        else:
            assert "count50" not in plans_allowed

    new_plan_added = option in ("add_plan", "add_plan_device")
    new_device_added = option in ("add_device", "add_plan_device")

    verify_allowed_lists(new_plan_added=new_plan_added, new_device_added=new_device_added)

    if new_device_added:
        assert devices_allowed_uid2 != devices_allowed_uid1
    else:
        assert devices_allowed_uid2 == devices_allowed_uid1
    if new_plan_added:
        assert plans_allowed_uid2 != plans_allowed_uid1
    else:
        assert plans_allowed_uid2 == plans_allowed_uid1

    # Reload the list of existing plans and devices from disk and make sure the new device/plan
    #   is loaded/not loaded depending on the update mode.
    resp6, _ = zmq_single_request("permissions_reload", params={"reload_plans_devices": True})
    assert resp6["success"] is True

    new_plan_added = new_plan_added and update_existing_plans_devices != "NEVER"
    new_device_added = new_device_added and update_existing_plans_devices != "NEVER"

    verify_allowed_lists(new_plan_added=new_plan_added, new_device_added=new_device_added)


_permissions_dict_not_allow_count = {
    "user_groups": {
        "root": {"allowed_plans": [None], "allowed_devices": [None]},
        "admin": {"allowed_plans": [None], "forbidden_plans": ["^count$"], "allowed_devices": [None]},
    }
}


# fmt: off
@pytest.mark.parametrize("sim_corrupt_redis_key", [False, True])
@pytest.mark.parametrize("user_group_permissions_reload", ["NEVER", "ON_REQUEST", "ON_STARTUP"])
# fmt: on
def test_cli_user_group_permissions_reload_01(
    re_manager_cmd, user_group_permissions_reload, sim_corrupt_redis_key  # noqa: F811
):
    """
    Tests for parameter ``--user-group-permissions-reload``: start manager, set permissions that
    are different from default (on disk), stop the manager and start it again (without removing
    Redis keys), check if correct permissions are loaded at startup, try to reload permissions
    using ``permissions_reload`` API and check if correct permissions are loaded.

    In addition, the case when Redis does not contain valid permissions is simulated.
    Permissions should be loaded from disk with any value of the parameter.
    """
    re = re_manager_cmd()

    resp1, _ = zmq_single_request(
        "permissions_set", params={"user_group_permissions": _permissions_dict_not_allow_count}
    )
    assert resp1["success"] is True, pprint.pformat(resp1)
    assert resp1["msg"] == ""

    # Stop the manager
    re.stop_manager(cleanup=False)

    # Simulate corrupt or missing redis key (should be treated the same).
    if sim_corrupt_redis_key:
        clear_redis_pool()

    # Start the manager again
    params = ["--user-group-permissions-reload", user_group_permissions_reload]
    re.start_manager(params=params, cleanup=False)

    resp2, _ = zmq_single_request("permissions_get")
    assert resp2["success"] is True
    user_group_permissions = resp2["user_group_permissions"]

    if (user_group_permissions_reload == "ON_STARTUP") or sim_corrupt_redis_key:
        assert user_group_permissions != _permissions_dict_not_allow_count
    else:
        assert user_group_permissions == _permissions_dict_not_allow_count

    # This should have no effect on the results of the test, but let's still check if it works.
    if sim_corrupt_redis_key:
        clear_redis_pool()

    # Attempt to reload permissions (from disk)
    resp3, _ = zmq_single_request("permissions_reload")

    if user_group_permissions_reload == "NEVER":
        assert resp3["success"] is False
        assert "RE Manager was started with option user_group_permissions_reload='NEVER'" in resp3["msg"]
    else:
        assert resp3["success"] is True

    resp4, _ = zmq_single_request("permissions_get")
    assert resp4["success"] is True
    user_group_permissions = resp4["user_group_permissions"]

    if (user_group_permissions_reload != "NEVER") or sim_corrupt_redis_key:
        assert user_group_permissions != _permissions_dict_not_allow_count
    else:
        assert user_group_permissions == _permissions_dict_not_allow_count



import asyncio
import pytest
import json
import copy
import re
import pprint
from bluesky_queueserver.manager.plan_queue_ops import PlanQueueOperations


errmsg_wrong_plan_type = "Parameter 'item' should be a dictionary"


class PQ:
    """
    The class that creates and initializes an instance of plan queue for testing and performs necessary
    cleanup afterwards. Intended for use with ``async with``.
    """

    def __init__(self):
        self._pq = None

    async def __aenter__(self):
        """
        Returns initialized instance of plan queue.
        """
        self._pq = PlanQueueOperations()
        await self._pq.start()
        # Clear any pool entries
        await self._pq.delete_pool_entries()
        await self._pq.user_group_permissions_clear()
        return self._pq

    async def __aexit__(self, exc_t, exc_v, exc_tb):
        """
        Cleanup after test.
        """
        # Don't leave any test entries in the pool
        await self._pq.delete_pool_entries()
        await self._pq.user_group_permissions_clear()


# fmt: off
@pytest.mark.parametrize("item_in, item_out", [
    ({"name": "plan1", "item_uid": "abcde"}, {"name": "plan1", "item_uid": "abcde"}),
    ({"user": "user1", "user_group": "group1"}, {"user": "user1", "user_group": "group1"}),
    ({"args": [1, 2], "kwargs": {"a": 1, "b": 2}}, {"args": [1, 2], "kwargs": {"a": 1, "b": 2}}),
    ({"item_type": "plan", "meta": {"md1": 1, "md2": 2}}, {"item_type": "plan", "meta": {"md1": 1, "md2": 2}}),
    ({"name": "plan1", "result": {}}, {"name": "plan1"}),
])
# fmt: on
def test_filter_item_parameters(item_in, item_out):
    """
    Tests for ``filter_item_parameters``.
    """

    async def testing():
        async with PQ() as pq:
            item = pq.filter_item_parameters(item_in)
            assert item == item_out

    asyncio.run(testing())


def test_running_plan_info():
    """
    Basic test for the following methods:
    `PlanQueueOperations.is_item_running()`
    `PlanQueueOperations.get_running_item_info()`
    `PlanQueueOperations.delete_pool_entries()`
    """

    async def testing():
        async with PQ() as pq:
            assert await pq.get_running_item_info() == {}
            assert await pq.is_item_running() is False

            some_plan = {"some_key": "some_value"}
            await pq._set_running_item_info(some_plan)
            assert await pq.get_running_item_info() == some_plan

            assert await pq.is_item_running() is True

            await pq._clear_running_item_info()
            assert await pq.get_running_item_info() == {}
            assert await pq.is_item_running() is False

            await pq._set_running_item_info(some_plan)
            await pq.delete_pool_entries()
            assert await pq.get_running_item_info() == {}
            assert await pq.is_item_running() is False

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("plan_running, plans, result_running, result_plans", [
    ({"testing": 1}, [{"testing": 2}, {"item_uid": "ab", "name": "nm"}, {"testing": 2}],
     {}, [{"item_uid": "ab", "name": "nm"}]),
    ({"testing": 1}, [{"testing": 2}, {"item_uid": "ab", "name": "nm"}, {"testing": 3}],
     {}, [{"item_uid": "ab", "name": "nm"}]),
    ({"item_uid": "a"}, [{"item_uid": "a1"}, {"item_uid": "a2"}, {"item_uid": "a3"}],
     {"item_uid": "a"}, [{"item_uid": "a1"}, {"item_uid": "a2"}, {"item_uid": "a3"}]),
])
# fmt: on
def test_queue_clean(plan_running, plans, result_running, result_plans):
    """
    Test for ``_queue_clean()`` method
    """

    async def testing():
        async with PQ() as pq:
            await pq._set_running_item_info(plan_running)
            for plan in plans:
                await pq._r_pool.rpush(pq._name_plan_queue, json.dumps(plan))

            assert await pq.get_running_item_info() == plan_running
            plan_queue, _ = await pq.get_queue()
            assert plan_queue == plans

            await pq._queue_clean()

            assert await pq.get_running_item_info() == result_running
            plan_queue, _ = await pq.get_queue()
            assert plan_queue == result_plans

    asyncio.run(testing())


@pytest.mark.parametrize("update", [False, True])
def test_set_plan_queue_mode_1(update):
    """
    Test basic functionality of ``set_plan_queue_mode`` function:
    The case: ``update=False``.
    """

    async def testing():
        async with PQ() as pq:
            # Initially plan queue mode must be default queue mode
            assert pq.plan_queue_mode == pq.plan_queue_mode_default
            assert pq.plan_queue_mode["loop"] is False

            # The properties are expecte to return copies
            assert pq.plan_queue_mode is not pq._plan_queue_mode
            assert pq.plan_queue_mode_default is not pq._plan_queue_mode_default

            queue_mode = {"loop": True}
            await pq.set_plan_queue_mode(queue_mode, update=update)
            assert pq._plan_queue_mode is not queue_mode  # Verify that 'set' operation performs copy
            assert pq.plan_queue_mode == queue_mode

            with pytest.raises(ValueError, match="Unsupported plan queue mode parameter 'nonexisting_key'"):
                await pq.set_plan_queue_mode({"nonexisting_key": True}, update=update)

            if update:
                await pq.set_plan_queue_mode({}, update=update)
            else:
                with pytest.raises(ValueError, match="Parameters {'loop'} are missing"):
                    await pq.set_plan_queue_mode({}, update=update)

            with pytest.raises(TypeError, match="Unsupported type .* of the parameter 'loop'"):
                await pq.set_plan_queue_mode({"loop": 10}, update=update)

            with pytest.raises(TypeError, match="Unsupported type .* of the parameter 'loop'"):
                await pq.set_plan_queue_mode({"loop": "some_string"}, update=update)

            # Verify that the queue mode is saved to Redis
            pq2 = PlanQueueOperations()
            await pq2.start()
            assert pq2.plan_queue_mode == queue_mode

            # Set queue mode to default
            assert pq.plan_queue_mode != pq.plan_queue_mode_default
            await pq.set_plan_queue_mode("default", update=update)
            assert pq.plan_queue_mode == pq.plan_queue_mode_default

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("plan, result",
                         [({"a": 10}, True),
                          ([10, 20], False),
                          (50, False),
                          ("abc", False)])
# fmt: on
def test_verify_item_type(plan, result):
    async def testing():
        async with PQ() as pq:
            if result:
                pq._verify_item_type(plan)
            else:
                with pytest.raises(TypeError, match=errmsg_wrong_plan_type):
                    pq._verify_item_type(plan)

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize(
    "plan, f_kwargs, result, errmsg",
    [({"a": 10}, {}, False, "Item does not have UID"),
     ([10, 20], {}, False, errmsg_wrong_plan_type),
     ({"item_uid": "one"}, {}, True, ""),
     ({"item_uid": "two"}, {}, False, "Item with UID .+ is already in the queue"),
     ({"item_uid": "three"}, {}, False, "Item with UID .+ is already in the queue"),
     ({"item_uid": "two"}, {"ignore_uids": None}, False, "Item with UID .+ is already in the queue"),
     ({"item_uid": "two"}, {"ignore_uids": ["two"]}, True, ""),
     ({"item_uid": "two"}, {"ignore_uids": ["two", "three"]}, True, ""),
     ({"item_uid": "two"}, {"ignore_uids": ["one", "three"]}, False, "Item with UID .+ is already in the queue"),
     ])
# fmt: on
def test_verify_item(plan, f_kwargs, result, errmsg):
    """
    Tests for method ``_verify_item()``.
    """
    # Set two existiing plans and then set one of them as running
    existing_plans = [{"item_type": "plan", "item_uid": "two"}, {"item_type": "plan", "item_uid": "three"}]

    async def testing():
        async with PQ() as pq:

            async def set_plans():
                # Add plan to queue
                for plan in existing_plans:
                    await pq.add_item_to_queue(plan)
                # Set one plan as currently running
                await pq.set_next_item_as_running()

                # Verify that setup is correct
                assert await pq.is_item_running() is True
                assert await pq.get_queue_size() == 1

            await set_plans()

            if result:
                pq._verify_item(plan, **f_kwargs)
            else:
                with pytest.raises(Exception, match=errmsg):
                    pq._verify_item(plan, **f_kwargs)

    asyncio.run(testing())


def test_new_item_uid():
    """
    Smoke test for the method ``new_item_uid()``.
    """
    assert isinstance(PlanQueueOperations.new_item_uid(), str)


# fmt: off
@pytest.mark.parametrize("plan", [
    {"name": "a"},
    {"item_uid": "some_uid", "name": "a"},
])
# fmt: on
def test_set_new_item_uuid(plan):
    """
    Basic test for the method ``set_new_item_uuid()``.
    """

    async def testing():
        async with PQ() as pq:
            uid = plan.get("item_uid", None)

            # The function is supposed to create or replace UID
            new_plan = pq.set_new_item_uuid(plan)

            assert "item_uid" in new_plan
            assert isinstance(new_plan["item_uid"], str)
            assert new_plan["item_uid"] != uid

    asyncio.run(testing())


def test_get_index_by_uid_1():
    """
    Test for ``_get_index_by_uid()``
    """
    plans = [
        {"item_uid": "a", "name": "name_a"},
        {"item_uid": "b", "name": "name_b"},
        {"item_uid": "c", "name": "name_c"},
    ]

    async def testing():
        async with PQ() as pq:
            for plan in plans:
                await pq.add_item_to_queue(plan)

            assert await pq._get_index_by_uid(uid="b") == 1

            with pytest.raises(IndexError, match="No plan with UID 'nonexistent'"):
                assert await pq._get_index_by_uid(uid="nonexistent")

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("sequence, indices_expected", [
    (["a", "b", "c"], [0, 1, 2]),
    (["a", "b"], [0, 1]),
    (["c", "b"], [2, 1]),
    (["c", "d", "b"], [2, -1, 1]),
])
# fmt: on
def test_get_index_by_uid_batch_1(sequence, indices_expected):
    """
    Test for ``_get_index_by_uid_batch()``
    """
    plans = [
        {"item_uid": "a", "name": "name_a"},
        {"item_uid": "b", "name": "name_b"},
        {"item_uid": "c", "name": "name_c"},
    ]

    async def testing():
        async with PQ() as pq:
            for plan in plans:
                await pq.add_item_to_queue(plan)

            indices = await pq._get_index_by_uid_batch(uids=sequence)
            assert indices == indices_expected

    asyncio.run(testing())


def test_uid_dict_1():
    """
    Basic test for functions associated with `_uid_dict`
    """
    plan_a = {"item_uid": "a", "name": "name_a"}
    plan_b = {"item_uid": "b", "name": "name_b"}
    plan_c = {"item_uid": "c", "name": "name_c"}

    plan_b_updated = {"item_uid": "b", "name": "name_b_updated"}

    async def testing():
        async with PQ() as pq:
            pq._uid_dict_add(plan_a)
            pq._uid_dict_add(plan_b)

            assert pq._is_uid_in_dict(plan_a["item_uid"]) is True
            assert pq._is_uid_in_dict(plan_b["item_uid"]) is True
            assert pq._is_uid_in_dict(plan_c["item_uid"]) is False

            assert pq._uid_dict_get_item(plan_b["item_uid"]) == plan_b
            pq._uid_dict_update(plan_b_updated)
            assert pq._uid_dict_get_item(plan_b["item_uid"]) == plan_b_updated

            pq._uid_dict_remove(plan_a["item_uid"])
            assert pq._is_uid_in_dict(plan_a["item_uid"]) is False
            assert pq._is_uid_in_dict(plan_b["item_uid"]) is True

            pq._uid_dict_clear()
            assert pq._is_uid_in_dict(plan_a["item_uid"]) is False
            assert pq._is_uid_in_dict(plan_b["item_uid"]) is False

    asyncio.run(testing())


def test_uid_dict_2():
    """
    Test if functions changing `pq._uid_dict` are also updating `pq.plan_queue_uid`.
    """
    plan_a = {"item_uid": "a", "name": "name_a"}
    plan_a_updated = {"item_uid": "a", "name": "name_a_updated"}

    async def testing():
        async with PQ() as pq:
            pq_uid = pq.plan_queue_uid

            pq._uid_dict_add(plan_a)

            assert pq.plan_queue_uid != pq_uid
            pq_uid = pq.plan_queue_uid

            pq._uid_dict_update(plan_a_updated)

            assert pq.plan_queue_uid != pq_uid
            pq_uid = pq.plan_queue_uid

            pq._uid_dict_remove(plan_a_updated["item_uid"])

            assert pq.plan_queue_uid != pq_uid
            pq_uid = pq.plan_queue_uid

            pq._uid_dict_clear()

            assert pq.plan_queue_uid != pq_uid

    asyncio.run(testing())


def test_uid_dict_3_initialize():
    """
    Basic test for functions associated with ``_uid_dict_initialize()``
    """

    async def testing():
        async with PQ() as pq:
            await pq.add_item_to_queue({"name": "a"})
            await pq.add_item_to_queue({"name": "b"})
            await pq.add_item_to_queue({"name": "c"})
            plans, _ = await pq.get_queue()
            uid_dict = {_["item_uid"]: _ for _ in plans}

            pq_uid = pq.plan_queue_uid

            await pq._uid_dict_initialize()

            assert pq._uid_dict == uid_dict
            assert pq.plan_queue_uid != pq_uid

    asyncio.run(testing())


def test_uid_dict_4_failing():
    """
    Failing cases for functions associated with `_uid_dict`
    """
    plan_a = {"item_uid": "a", "name": "name_a"}
    plan_b = {"item_uid": "b", "name": "name_b"}
    plan_c = {"item_uid": "c", "name": "name_c"}

    async def testing():
        async with PQ() as pq:

            pq._uid_dict_add(plan_a)
            pq._uid_dict_add(plan_b)

            # Add plan with UID that already exists
            pq_uid = pq.plan_queue_uid
            with pytest.raises(RuntimeError, match=f"'{plan_a['item_uid']}', which is already in the queue"):
                pq._uid_dict_add(plan_a)
            assert pq.plan_queue_uid == pq_uid

            assert len(pq._uid_dict) == 2

            # Remove plan with UID does not exist exists
            with pytest.raises(RuntimeError, match=f"'{plan_c['item_uid']}', which is not in the queue"):
                pq._uid_dict_remove(plan_c["item_uid"])
            assert pq.plan_queue_uid == pq_uid

            assert len(pq._uid_dict) == 2

            # Update plan with UID does not exist exists
            with pytest.raises(RuntimeError, match=f"'{plan_c['item_uid']}', which is not in the queue"):
                pq._uid_dict_update(plan_c)
            assert pq.plan_queue_uid == pq_uid

            assert len(pq._uid_dict) == 2

    asyncio.run(testing())


def test_remove_item():
    """
    Basic test for functions associated with ``_remove_plan()``
    """

    async def testing():
        async with PQ() as pq:

            plan_list = [{"name": "a"}, {"name": "b"}, {"name": "c"}]
            for plan in plan_list:
                await pq.add_item_to_queue(plan)

            plans, _ = await pq.get_queue()
            plan_to_remove = [_ for _ in plans if _["name"] == "b"][0]

            # Remove one plan
            await pq._remove_item(plan_to_remove)
            plans, _ = await pq.get_queue()
            assert len(plans) == 2

            # Add a copy of a plan (queue is not supposed to have copies in real life)
            plan_to_add = plans[0]
            await pq._r_pool.lpush(pq._name_plan_queue, json.dumps(plan_to_add))
            # Now remove both plans
            await pq._remove_item(plan_to_add, single=False)  # Allow deleting multiple or no plans
            assert await pq.get_queue_size() == 1

            # Delete the plan again (the plan is not in the queue, but it shouldn't raise an exception)
            await pq._remove_item(plan_to_add, single=False)  # Allow deleting multiple or no plans
            assert await pq.get_queue_size() == 1

            with pytest.raises(RuntimeError, match="One item is expected"):
                await pq._remove_item(plan_to_add)
            assert await pq.get_queue_size() == 1

            # Now add 'plan_to_add' twice (create two copies)
            await pq._r_pool.lpush(pq._name_plan_queue, json.dumps(plan_to_add))
            await pq._r_pool.lpush(pq._name_plan_queue, json.dumps(plan_to_add))
            assert await pq.get_queue_size() == 3
            # Attempt to delete two copies
            with pytest.raises(RuntimeError, match="One item is expected"):
                await pq._remove_item(plan_to_add)
            # Exception is raised, but both copies are deleted
            assert await pq.get_queue_size() == 1

    asyncio.run(testing())


def test_get_queue_full_1():
    """
    Basic test for the functions ``PlanQueueOperations.get_queue()`` and
    ``PlanQueueOperations.get_queue_full()``
    """

    async def testing():
        async with PQ() as pq:
            plans = [
                {"item_type": "plan", "item_uid": "one", "name": "a"},
                {"item_type": "plan", "item_uid": "two", "name": "b"},
                {"item_type": "plan", "item_uid": "three", "name": "c"},
            ]

            for p in plans:
                await pq.add_item_to_queue(p)
            await pq.set_next_item_as_running()

            pq_uid = pq.plan_queue_uid
            queue1, uid1 = await pq.get_queue()
            running_item1 = await pq.get_running_item_info()
            queue2, running_item2, uid2 = await pq.get_queue_full()

            assert queue1 == plans[1:]
            assert queue2 == plans[1:]
            assert running_item1 == plans[0]
            assert running_item2 == plans[0]
            assert uid1 == pq_uid
            assert uid2 == pq_uid

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("params, name", [
    ({"pos": "front"}, "a"),
    ({"pos": "back"}, "c"),
    ({"pos": 0}, "a"),
    ({"pos": 1}, "b"),
    ({"pos": 2}, "c"),
    ({"pos": 3}, None),  # Index out of range
    ({"pos": -1}, "c"),
    ({"pos": -2}, "b"),
    ({"pos": -3}, "a"),
    ({"pos": -4}, None),  # Index out of range
    ({"uid": "one"}, "a"),
    ({"uid": "two"}, "b"),
    ({"uid": "nonexistent"}, None),
])
# fmt: on
def test_get_item_1(params, name):
    """
    Basic test for the function ``PlanQueueOperations.get_item()``
    """

    async def testing():
        async with PQ() as pq:
            pq_uid = pq.plan_queue_uid
            await pq.add_item_to_queue({"item_uid": "one", "name": "a"})
            await pq.add_item_to_queue({"item_uid": "two", "name": "b"})
            await pq.add_item_to_queue({"item_uid": "three", "name": "c"})
            assert await pq.get_queue_size() == 3
            assert pq.plan_queue_uid != pq_uid

            if name is not None:
                plan = await pq.get_item(**params)
                assert plan["name"] == name
            else:
                msg = "Index .* is out of range" if "pos" in params else "is not in the queue"
                with pytest.raises(IndexError, match=msg):
                    await pq.get_item(**params)

    asyncio.run(testing())


def test_get_item_2_fail():
    """
    Basic test for the function ``PlanQueueOperations.get_item()``.
    Attempt to retrieve a running plan.
    """

    async def testing():
        async with PQ() as pq:
            await pq.add_item_to_queue({"item_type": "plan", "item_uid": "one", "name": "a"})
            await pq.add_item_to_queue({"item_type": "plan", "item_uid": "two", "name": "b"})
            await pq.add_item_to_queue({"item_type": "plan", "item_uid": "three", "name": "c"})
            assert await pq.get_queue_size() == 3

            pq_uid = pq.plan_queue_uid
            await pq.set_next_item_as_running()
            assert await pq.get_queue_size() == 2
            assert pq.plan_queue_uid != pq_uid

            with pytest.raises(IndexError, match="is currently running"):
                await pq.get_item(uid="one")

            # Ambiguous parameters (position and UID is passed)
            with pytest.raises(ValueError, match="Ambiguous parameters"):
                await pq.get_item(pos=5, uid="abc")

    asyncio.run(testing())


def test_add_item_to_queue_1():
    """
    Basic test for the function ``PlanQueueOperations.add_item_to_queue()``
    """

    async def testing():
        async with PQ() as pq:

            async def add_plan(plan, n, **kwargs):
                plan_added, qsize = await pq.add_item_to_queue(plan, **kwargs)
                assert plan_added["name"] == plan["name"], f"plan: {plan}"
                assert qsize == n, f"plan: {plan}"

            await add_plan({"name": "a"}, 1)
            await add_plan({"name": "b"}, 2)
            await add_plan({"name": "c"}, 3, pos="back")
            await add_plan({"name": "d"}, 4, pos="front")
            await add_plan({"name": "e"}, 5, pos=0)  # front
            await add_plan({"name": "f"}, 6, pos=5)  # back (index == queue size)
            await add_plan({"name": "g"}, 7, pos=5)  # previous to last
            await add_plan({"name": "h"}, 8, pos=-1)  # previous to last
            await add_plan({"name": "i"}, 9, pos=3)  # arbitrary index
            await add_plan({"name": "j"}, 10, pos=100)  # back (index some large number)
            await add_plan({"name": "k"}, 11, pos=-10)  # front (precisely negative queue size)
            await add_plan({"name": "l"}, 12, pos=-100)  # front (index some large negative number)

            assert await pq.get_queue_size() == 12

            plans, _ = await pq.get_queue()
            name_sequence = [_["name"] for _ in plans]
            assert name_sequence == ["l", "k", "e", "d", "a", "i", "b", "c", "g", "h", "f", "j"]

            await pq.clear_queue()

    asyncio.run(testing())


def test_add_item_to_queue_2():
    """
    Basic test for the function ``PlanQueueOperations.add_item_to_queue()``
    """

    async def testing():
        async with PQ() as pq:

            async def add_plan(plan, n, **kwargs):
                plan_added, qsize = await pq.add_item_to_queue(plan, **kwargs)
                assert plan_added["name"] == plan["name"], f"plan: {plan}"
                assert qsize == n, f"plan: {plan}"

            await add_plan({"item_type": "plan", "name": "a"}, 1)
            await add_plan({"item_type": "plan", "name": "b"}, 2)
            await add_plan({"item_type": "plan", "name": "c"}, 3, pos="back")

            plan_queue, _ = await pq.get_queue()
            displaced_uid = plan_queue[1]["item_uid"]

            await add_plan({"item_type": "plan", "name": "d"}, 4, before_uid=displaced_uid)
            await add_plan({"item_type": "plan", "name": "e"}, 5, after_uid=displaced_uid)

            # This reduces the number of elements in the queue by one
            await pq.set_next_item_as_running()

            displaced_uid = plan_queue[0]["item_uid"]
            await add_plan({"name": "f"}, 5, after_uid=displaced_uid)

            with pytest.raises(
                IndexError, match="Can not insert a plan in the queue before a currently running plan"
            ):
                await add_plan({"name": "g"}, 5, before_uid=displaced_uid)

            with pytest.raises(IndexError, match="is not in the queue"):
                await add_plan({"name": "h"}, 5, before_uid="nonexistent_uid")

            assert await pq.get_queue_size() == 5

            plans, _ = await pq.get_queue()
            name_sequence = [_["name"] for _ in plans]
            assert name_sequence == ["f", "d", "b", "e", "c"]

            await pq.clear_queue()

    asyncio.run(testing())


@pytest.mark.parametrize("filter_params", [False, True])
def test_add_item_to_queue_3(filter_params):
    """
    Test if parameter filtering works as expected with `add_item_to_queue` function.
    """

    async def testing():
        async with PQ() as pq:
            # Parameter 'result' should be removed if filtering is enabled
            plan1 = {"item_type": "plan", "name": "a", "item_uid": "1"}
            plan2 = plan1.copy()
            plan2["result"] = {}

            plan_added, qsize = await pq.add_item_to_queue(plan2, filter_parameters=filter_params)

            if filter_params:
                assert plan_added == plan1
            else:
                assert plan_added == plan2

    asyncio.run(testing())


def test_add_item_to_queue_4_fail():
    """
    Failing tests for the function ``PlanQueueOperations.add_item_to_queue()``
    """

    async def testing():
        async with PQ() as pq:
            pq_uid = pq.plan_queue_uid
            with pytest.raises(ValueError, match="Parameter 'pos' has incorrect value"):
                await pq.add_item_to_queue({"name": "a"}, pos="something")
            assert pq.plan_queue_uid == pq_uid

            with pytest.raises(TypeError, match=errmsg_wrong_plan_type):
                await pq.add_item_to_queue("plan_is_not_string")
            assert pq.plan_queue_uid == pq_uid

            # Duplicate plan UID
            plan = {"item_uid": "abc", "name": "a"}
            await pq.add_item_to_queue(plan)
            pq_uid = pq.plan_queue_uid
            with pytest.raises(RuntimeError, match="Item with UID .+ is already in the queue"):
                await pq.add_item_to_queue(plan)
            assert pq.plan_queue_uid == pq_uid

            # Ambiguous parameters (position and UID is passed)
            with pytest.raises(ValueError, match="Ambiguous parameters"):
                await pq.add_item_to_queue({"name": "abc"}, pos=5, before_uid="abc")
            assert pq.plan_queue_uid == pq_uid

            # Ambiguous parameters ('before_uid' and 'after_uid' is specified)
            with pytest.raises(ValueError, match="Ambiguous parameters"):
                await pq.add_item_to_queue({"name": "abc"}, before_uid="abc", after_uid="abc")
            assert pq.plan_queue_uid == pq_uid

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("batch_params, queue_seq, batch_seq, expected_seq", [
    ({}, "", "", ""),  # Add an empty batch
    ({}, "", "567", "567"),
    ({"pos": "front"}, "", "567", "567"),
    ({"pos": "back"}, "", "567", "567"),
    ({}, "1234", "567", "1234567"),
    ({"pos": "front"}, "1234", "567", "5671234"),
    ({"pos": "back"}, "1234", "567", "1234567"),
    ({"pos": 0}, "1234", "567", "5671234"),
    ({"pos": 1}, "1234", "567", "1567234"),
    ({"pos": 100}, "1234", "567", "1234567"),
    ({"pos": -1}, "1234", "567", "1235674"),
    ({"pos": -100}, "1234", "567", "5671234"),
    ({"before_uid": "1"}, "1234", "567", "5671234"),
    ({"before_uid": "2"}, "1234", "567", "1567234"),
    ({"before_uid": "3"}, "1234", "567", "1256734"),
    ({"after_uid": "1"}, "1234", "567", "1567234"),
    ({"after_uid": "2"}, "1234", "567", "1256734"),
    ({"after_uid": "4"}, "1234", "567", "1234567"),
])
# fmt: on
def test_add_batch_to_queue_1(batch_params, queue_seq, batch_seq, expected_seq):
    """
    Basic test for the function ``PlanQueueOperations.add_batch_to_queue()``
    """

    async def testing():
        async with PQ() as pq:

            async def add_plan(plan, n, **kwargs):
                plan_added, qsize = await pq.add_item_to_queue(plan, **kwargs)
                assert plan_added["name"] == plan["name"], f"plan: {plan}"
                assert qsize == n, f"plan: {plan}"

            # Create the queue with plans
            for n, p_name in enumerate(queue_seq):
                await add_plan({"name": p_name, "item_uid": f"{p_name}{p_name}"}, n + 1)

            def fix_uid(uid):
                return f"{uid}{uid}"

            if "before_uid" in batch_params:
                batch_params["before_uid"] = fix_uid(batch_params["before_uid"])
            if "after_uid" in batch_params:
                batch_params["after_uid"] = fix_uid(batch_params["after_uid"])

            items = []
            for p_name in batch_seq:
                items.append({"name": p_name, "item_uid": p_name + p_name})
            items_added, results, qsize, success = await pq.add_batch_to_queue(items, **batch_params)
            assert success is True, pprint.pformat(results)
            assert qsize == len(queue_seq) + len(batch_seq)
            assert await pq.get_queue_size() == len(queue_seq) + len(batch_seq)
            assert len(items_added) == len(items)
            assert len(results) == len(items)

            # Verify that the results are set correctly (success)
            for res in results:
                assert res["success"] is True, pprint.pformat(results)
                assert res["msg"] == "", pprint.pformat(results)

            # Verify the sequence of items in the queue
            queue, _ = await pq.get_queue()
            queue_sequence = [_["name"] for _ in queue]
            queue_sequence = "".join(queue_sequence)
            assert queue_sequence == expected_seq

            await pq.clear_queue()

    asyncio.run(testing())


@pytest.mark.parametrize("filter_params", [False, True, None])
def test_add_batch_to_queue_2(filter_params):
    """
    Test if parameter filtering works as expected with `add_batch_to_queue` function.
    """

    async def testing():
        async with PQ() as pq:
            # Parameter 'result' should be removed if filtering is enabled.
            params = {"filter_parameters": filter_params} if (filter_params is not None) else {}
            do_filtering = True if (filter_params is None) else filter_params

            items = [{"name": _, "result": {}} for _ in ("a", "b", "c")]
            items_added, results, qsize, success = await pq.add_batch_to_queue(items, **params)
            assert success is True
            assert qsize == len(items)
            assert len(items_added) == len(items)
            assert len(results) == len(items)

            queue, _ = await pq.get_queue()

            assert [_["name"] for _ in queue] == [_["name"] for _ in items]
            for queue_item in queue:
                if do_filtering:
                    assert "result" not in queue_item, pprint.pformat(queue_item)
                else:
                    assert "result" in queue_item, pprint.pformat(queue_item)

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("params, queue_seq, batch_seq, err_msgs", [
    ({}, "abcd", "eaf", ["", "Item with UID .+ is already in the queue", ""]),
    ({}, "abcd", "ead", [""] + ["Item with UID .+ is already in the queue"] * 2),
    ({"before_uid": "unknown"}, "abcd", "efg", ["Plan with UID .* is not in the queue"] * 3),
    ({"after_uid": "unknown"}, "abcd", "efg", ["Plan with UID .* is not in the queue"] * 3),
    ({"before_uid": "unknown", "after_uid": "unknown"}, "abcd", "efg", ["Ambiguous parameters"] * 3),
    ({"pos": "front", "after_uid": "unknown"}, "abcd", "efg", ["Ambiguous parameters"] * 3),
])
# fmt: on
def test_add_batch_to_queue_3_fail(params, queue_seq, batch_seq, err_msgs):
    """
    Failing cases for the function ``PlanQueueOperations.add_batch_to_queue()``
    """

    async def testing():
        async with PQ() as pq:

            async def add_plan(plan, n, **kwargs):
                plan_added, qsize = await pq.add_item_to_queue(plan, **kwargs)
                assert plan_added["name"] == plan["name"], f"plan: {plan}"
                assert qsize == n, f"plan: {plan}"

            # Create the queue with plans
            for n, p_name in enumerate(queue_seq):
                await add_plan({"name": p_name, "item_uid": p_name + p_name}, n + 1)

            items = []
            for p_name in batch_seq:
                items.append({"name": p_name, "item_uid": p_name + p_name})
            items_added, results, qsize, success = await pq.add_batch_to_queue(items, **params)

            assert success is False
            assert qsize == len(queue_seq)
            assert len(items_added) == len(items)
            assert len(results) == len(items)

            # The returned item list is expected to be identical to the original list
            assert items_added == items, pprint.pformat(items_added)

            for res, msg in zip(results, err_msgs):
                if not msg:
                    assert res["success"] is True, pprint.pformat(res)
                    assert res["msg"] == "", pprint.pformat(res)
                else:
                    assert res["success"] is False, pprint.pformat(res)
                    assert re.search(msg, res["msg"]), pprint.pformat(res)

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("replace_uid", [False, True])
# fmt: on
def test_replace_item_1(replace_uid):
    """
    Basic functionality of ``PlanQueueOperations.replace_item()`` function.
    """

    async def testing():
        async with PQ() as pq:
            plans = [{"name": "a"}, {"name": "b"}, {"name": "c"}]
            plans_added = [None] * len(plans)
            qsizes = [None] * len(plans)

            for n, plan in enumerate(plans):
                plans_added[n], qsizes[n] = await pq.add_item_to_queue(plan)

            assert qsizes == [1, 2, 3]
            assert await pq.get_queue_size() == 3

            # Change name, but keep UID
            plan_names_new = ["d", "e", "f"]
            for n in range(len(plans_added)):
                plan = plans_added[n].copy()
                plan["name"] = plan_names_new[n]

                uid_to_replace = plan["item_uid"]
                if replace_uid:
                    plan["item_uid"] = pq.new_item_uid()  # Generate new UID

                pq_uid = pq.plan_queue_uid
                plan_new, qsize = await pq.replace_item(plan, item_uid=uid_to_replace)
                assert plan_new["name"] == plan["name"]
                assert plan_new["item_uid"] == plan["item_uid"]
                assert pq.plan_queue_uid != pq_uid

                assert await pq.get_queue_size() == 3
                plans_added[n] = plan_new

                # Make sure that the plan can be correctly extracted by uid
                assert await pq.get_item(uid=plan["item_uid"]) == plan
                assert await pq.get_queue_size() == 3

                # Initialize '_uid_dict' and see if the plan can still be extracted using correct UID.
                await pq._uid_dict_initialize()
                assert await pq.get_item(uid=plan["item_uid"]) == plan
                assert await pq.get_queue_size() == 3

    asyncio.run(testing())


def test_replace_item_2():
    """
    ``PlanQueueOperations.replace_item()`` function: not UID in the plan - random UID is assigned.
    """

    async def testing():
        async with PQ() as pq:
            plans = [{"name": "a"}, {"name": "b"}, {"name": "c"}]
            plans_added = [None] * len(plans)
            qsizes = [None] * len(plans)

            for n, plan in enumerate(plans):
                plans_added[n], qsizes[n] = await pq.add_item_to_queue(plan)

            assert qsizes == [1, 2, 3]
            assert await pq.get_queue_size() == 3

            new_name = "h"
            plan_new = {"name": new_name}  # No UID in the plan. It should still be inserted
            pq_uid = pq.plan_queue_uid
            plan_replaced, qsize = await pq.replace_item(plan_new, item_uid=plans_added[1]["item_uid"])
            assert pq.plan_queue_uid != pq_uid

            new_uid = plan_replaced["item_uid"]
            assert new_uid != plans_added[1]["item_uid"]
            assert plan_replaced["name"] == plan_new["name"]

            plan = await pq.get_item(uid=new_uid)
            assert plan["item_uid"] == new_uid
            assert plan["name"] == new_name

            # Initialize '_uid_dict' and see if the plan can still be extracted using correct UID.
            await pq._uid_dict_initialize()
            plan = await pq.get_item(uid=new_uid)
            assert plan["item_uid"] == new_uid
            assert plan["name"] == new_name

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("mode", ["exact_copy", "change_uid", "change_plan"])
# fmt: on
def test_replace_item_3(mode):
    """
    ``PlanQueueOperations.replace_item()`` function: make sure that filtering
      is applied to the parameters of edited plan if the plan was changed
    """

    async def testing():
        async with PQ() as pq:
            plans = [{"name": "a", "result": {}}, {"name": "b"}, {"name": "c"}]

            for n, plan in enumerate(plans):
                await pq.add_item_to_queue(plan, filter_parameters=False)

            assert await pq.get_queue_size() == 3

            queue, _ = await pq.get_queue()
            plan0_before = queue[0]
            assert "result" in plan0_before

            if mode == "exact_copy":
                # Replace the plan with the exact copy of it. Filtering is not be applied.
                plan_new = plan0_before
                plan_replaced, qsize = await pq.replace_item(plan_new, item_uid=plan_new["item_uid"])

                queue, _ = await pq.get_queue()
                plan0_after = queue[0]

                assert plan_replaced == plan0_after
                assert "result" in plan_replaced, pprint.pformat(plan_replaced)

            elif mode == "change_uid":
                # Modify plan UID. Filtering is applied
                plan_new = plan0_before
                item_uid_to_replace = plan_new["item_uid"]
                plan_new = pq.set_new_item_uuid(plan_new)
                plan_replaced, qsize = await pq.replace_item(plan_new, item_uid=item_uid_to_replace)

                queue, _ = await pq.get_queue()
                plan0_after = queue[0]

                assert plan_replaced == plan0_after
                assert "result" not in plan_replaced, pprint.pformat(plan_replaced)

            elif mode == "change_plan":
                # Modify plan, keep UID the same. Filtering is applied
                plan_new = plan0_before
                new_name = "h"
                plan_new["name"] = new_name
                plan_replaced, qsize = await pq.replace_item(plan_new, item_uid=plan_new["item_uid"])

                queue, _ = await pq.get_queue()
                plan0_after = queue[0]

                assert plan_replaced == plan0_after
                assert "result" not in plan_replaced, pprint.pformat(plan_replaced)

            else:
                raise Exception(f"Test mode '{mode}' does not exist.")

    asyncio.run(testing())


def test_replace_item_4_failing():
    """
    ``PlanQueueOperations.replace_item()`` - failing cases
    """

    async def testing():
        async with PQ() as pq:
            plans = [
                {"item_type": "plan", "name": "a"},
                {"item_type": "plan", "name": "b"},
                {"item_type": "plan", "name": "c"},
            ]
            plans_added = [None] * len(plans)
            qsizes = [None] * len(plans)

            for n, plan in enumerate(plans):
                plans_added[n], qsizes[n] = await pq.add_item_to_queue(plan)

            assert qsizes == [1, 2, 3]
            assert await pq.get_queue_size() == 3

            # Set the first item as 'running'
            running_plan = await pq.set_next_item_as_running()
            assert running_plan == plans_added[0]

            queue, _ = await pq.get_queue()
            running_item_info = await pq.get_running_item_info()

            pq_uid = pq.plan_queue_uid

            plan_new = {"name": "h"}  # No UID in the plan. It should still be inserted
            # Case: attempt to replace a plan that is currently running
            with pytest.raises(
                RuntimeError, match="Failed to replace item: Item with UID .* is currently running"
            ):
                await pq.replace_item(plan_new, item_uid=running_plan["item_uid"])
            assert pq.plan_queue_uid == pq_uid

            # Case: attempt to replace a plan that is not in queue
            with pytest.raises(RuntimeError, match="Failed to replace item: Item with UID .* is not in the queue"):
                await pq.replace_item(plan_new, item_uid="uid-that-does-not-exist")
            assert pq.plan_queue_uid == pq_uid

            # Case: attempt to replace a plan with another plan that already exists in the queue
            plan = plans_added[1]
            with pytest.raises(RuntimeError, match="Item with UID .* is already in the queue"):
                await pq.replace_item(plan, item_uid=plans_added[2]["item_uid"])
            assert pq.plan_queue_uid == pq_uid

            # Case: attempt to replace a plan with currently running plan
            with pytest.raises(RuntimeError, match="Item with UID .* is already in the queue"):
                await pq.replace_item(running_plan, item_uid=plans_added[2]["item_uid"])
            assert pq.plan_queue_uid == pq_uid

            # Make sure that the queue did not change during the test
            plan_queue, _ = await pq.get_queue()
            assert plan_queue == queue
            assert await pq.get_running_item_info() == running_item_info

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("params, src, order, success, pquid_changed, msg", [
    ({"pos": 1, "pos_dest": 1}, 1, "abcde", True, False, ""),
    ({"pos": "front", "pos_dest": "front"}, 0, "abcde", True, False, ""),
    ({"pos": "back", "pos_dest": "back"}, 4, "abcde", True, False, ""),
    ({"pos": "front", "pos_dest": "back"}, 0, "bcdea", True, True, ""),
    ({"pos": "back", "pos_dest": "front"}, 4, "eabcd", True, True, ""),
    ({"pos": 1, "pos_dest": 2}, 1, "acbde", True, True, ""),
    ({"pos": 2, "pos_dest": 1}, 2, "acbde", True, True, ""),
    ({"pos": 0, "pos_dest": 4}, 0, "bcdea", True, True, ""),
    ({"pos": 4, "pos_dest": 0}, 4, "eabcd", True, True, ""),
    ({"pos": 3, "pos_dest": "front"}, 3, "dabce", True, True, ""),
    ({"pos": 2, "pos_dest": "back"}, 2, "abdec", True, True, ""),
    ({"uid": "p3", "after_uid": "p3"}, 2, "abcde", True, False, ""),
    ({"uid": "p1", "before_uid": "p2"}, 0, "abcde", True, True, ""),
    ({"uid": "p1", "after_uid": "p2"}, 0, "bacde", True, True, ""),
    ({"uid": "p2", "pos_dest": "front"}, 1, "bacde", True, True, ""),
    ({"uid": "p2", "pos_dest": "back"}, 1, "acdeb", True, True, ""),
    ({"uid": "p1", "pos_dest": "front"}, 0, "abcde", True, False, ""),
    ({"uid": "p5", "pos_dest": "back"}, 4, "abcde", True, False, ""),
    ({"pos": 1, "after_uid": "p4"}, 1, "acdbe", True, True, ""),
    ({"pos": "front", "after_uid": "p4"}, 0, "bcdae", True, True, ""),
    ({"pos": 3, "after_uid": "p1"}, 3, "adbce", True, True, ""),
    ({"pos": "back", "after_uid": "p1"}, 4, "aebcd", True, True, ""),
    ({"pos": 1, "before_uid": "p4"}, 1, "acbde", True, True, ""),
    ({"pos": "front", "before_uid": "p4"}, 0, "bcade", True, True, ""),
    ({"pos": 3, "before_uid": "p1"}, 3, "dabce", True, True, ""),
    ({"pos": "back", "before_uid": "p1"}, 4, "eabcd", True, True, ""),
    ({"pos": "back", "after_uid": "p5"}, 4, "abcde", True, False, ""),
    ({"pos": "front", "before_uid": "p1"}, 0, "abcde", True, False, ""),
    ({"pos": 50, "before_uid": "p1"}, 0, "", False, False, r"Source plan \(position 50\) was not found"),
    ({"uid": "abc", "before_uid": "p1"}, 0, "", False, False, r"Source plan \(UID 'abc'\) was not found"),
    ({"pos": 3, "pos_dest": 50}, 0, "", False, False, r"Destination plan \(position 50\) was not found"),
    ({"uid": "p1", "before_uid": "abc"}, 0, "", False, False, r"Destination plan \(UID 'abc'\) was not found"),
    ({"before_uid": "p1"}, 0, "", False, None, r"Source position or UID is not specified"),
    ({"pos": 3}, 0, "", False, False, r"Destination position or UID is not specified"),
    ({"pos": 1, "uid": "p1", "before_uid": "p4"}, 1, "", False, False, "Ambiguous parameters"),
    ({"pos": 1, "pos_dest": 4, "before_uid": "p4"}, 1, "", False, False, "Ambiguous parameters"),
    ({"pos": 1, "after_uid": "p4", "before_uid": "p4"}, 1, "", False, False, "Ambiguous parameters"),
])
# fmt: on
def test_move_item_1(params, src, order, success, pquid_changed, msg):
    """
    Basic tests for ``move_item()``.
    """

    async def testing():
        async with PQ() as pq:
            plans = [
                {"item_uid": "p1", "name": "a"},
                {"item_uid": "p2", "name": "b"},
                {"item_uid": "p3", "name": "c"},
                {"item_uid": "p4", "name": "d"},
                {"item_uid": "p5", "name": "e"},
            ]

            for plan in plans:
                await pq.add_item_to_queue(plan)

            assert await pq.get_queue_size() == len(plans)
            pq_uid = pq.plan_queue_uid

            if success:
                plan, qsize = await pq.move_item(**params)
                assert qsize == len(plans)
                assert plan["name"] == plans[src]["name"]

                queue, _ = await pq.get_queue()
                names = [_["name"] for _ in queue]
                names = "".join(names)
                assert names == order

                if pquid_changed:
                    assert pq.plan_queue_uid != pq_uid
                else:
                    assert pq.plan_queue_uid == pq_uid

            else:
                with pytest.raises(Exception, match=msg):
                    await pq.move_item(**params)
                assert pq.plan_queue_uid == pq_uid

    asyncio.run(testing())


def test_move_item_2():
    """
    ``move_item``: test if the item is moved 'as is', i.e. no parameter filtering is applied to it.
    """

    async def testing():
        async with PQ() as pq:
            plans = [
                {"item_uid": "p1", "name": "a", "result": {}},
                {"item_uid": "p2", "name": "b"},
                {"item_uid": "p3", "name": "c"},
            ]

            for plan in plans:
                await pq.add_item_to_queue(plan, filter_parameters=False)

            queue, _ = await pq.get_queue()
            assert await pq.get_queue_size() == 3
            assert "result" in queue[0]

            uid_src = queue[0]["item_uid"]
            uid_dest = queue[1]["item_uid"]

            await pq.move_item(uid=uid_src, after_uid=uid_dest)

            queue, _ = await pq.get_queue()
            assert await pq.get_queue_size() == 3
            # Make sure that the items were rearranged
            assert queue[0]["item_uid"] == uid_dest, pprint.pformat(queue)
            assert queue[1]["item_uid"] == uid_src, pprint.pformat(queue)
            # Make sure that the 'result' parameter was not removed
            assert "result" in queue[1]

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("batch_params, queue_seq, selection_seq, batch_seq, expected_seq, success, msg", [
    ({"pos_dest": "front"}, "0123456", "23", "23", "2301456", True, ""),
    ({"before_uid": "0"}, "0123456", "23", "23", "2301456", True, ""),
    ({"pos_dest": "back"}, "0123456", "23", "23", "0145623", True, ""),
    ({"after_uid": "6"}, "0123456", "23", "23", "0145623", True, ""),
    ({"before_uid": "5"}, "0123456", "23", "23", "0142356", True, ""),
    ({"after_uid": "5"}, "0123456", "23", "23", "0145236", True, ""),
    # Controlling the order of moved items
    ({"after_uid": "5"}, "0123456", "023", "023", "1450236", True, ""),
    ({"after_uid": "5"}, "0123456", "203", "203", "1452036", True, ""),
    ({"after_uid": "5"}, "0123456", "302", "302", "1453026", True, ""),
    ({"after_uid": "5", "reorder": False}, "0123456", "302", "302", "1453026", True, ""),
    ({"after_uid": "5", "reorder": True}, "0123456", "023", "023", "1450236", True, ""),
    ({"after_uid": "5", "reorder": True}, "0123456", "203", "023", "1450236", True, ""),
    ({"after_uid": "5", "reorder": True}, "0123456", "302", "023", "1450236", True, ""),
    # Empty list of UIDS
    ({"pos_dest": "front"}, "0123456", "", "", "0123456", True, ""),
    ({"pos_dest": "front"}, "", "", "", "", True, ""),
    # Move the batch which is already in the front or back to front or back of the queue
    #   (nothing is done, but operation is still successful)
    ({"pos_dest": "front"}, "0123456", "01", "01", "0123456", True, ""),
    ({"pos_dest": "back"}, "0123456", "56", "56", "0123456", True, ""),
    # Same, but change the order of moved items
    ({"pos_dest": "front"}, "0123456", "10", "10", "1023456", True, ""),
    ({"pos_dest": "back"}, "0123456", "65", "65", "0123465", True, ""),
    # Failing cases
    ({}, "0123456", "23", "23", "0123456", False, "Destination for the batch is not specified"),
    ({"pos_dest": "front", "before_uid": "5"}, "0123456", "23", "23", "0123456", False,
     "more than one mutually exclusive parameter"),
    ({"after_uid": "3"}, "0123456", "023", "023", "0123456", False, "item with UID '33' is in the batch"),
    ({"before_uid": "3"}, "0123456", "023", "023", "0123456", False, "item with UID '33' is in the batch"),
    ({"after_uid": "5"}, "0123456", "093", "093", "0123456", False,
     re.escape("The queue does not contain items with the following UIDs: ['99']")),
    ({"after_uid": "5"}, "0123456", "07893", "07893", "0123456", False,
     re.escape("The queue does not contain items with the following UIDs: ['77', '88', '99']")),
    ({"after_uid": "5"}, "0123456", "0223", "0223", "0123456", False,
     re.escape("The list of contains repeated UIDs (1 UIDs)")),
])
# fmt: on
def test_move_batch_1(batch_params, queue_seq, selection_seq, batch_seq, expected_seq, success, msg):
    async def testing():
        async with PQ() as pq:

            async def add_plan(plan, n, **kwargs):
                plan_added, qsize = await pq.add_item_to_queue(plan, **kwargs)
                assert plan_added["name"] == plan["name"], f"plan: {plan}"
                assert qsize == n, f"plan: {plan}"

            def name_to_uid(uid):
                return f"{uid}{uid}"

            # Create the queue with plans
            for n, p_name in enumerate(queue_seq):
                await add_plan({"name": p_name, "item_uid": f"{name_to_uid(p_name)}"}, n + 1)
            qsize = await pq.get_queue_size()
            assert qsize == len(queue_seq)

            if "before_uid" in batch_params:
                batch_params["before_uid"] = name_to_uid(batch_params["before_uid"])
            if "after_uid" in batch_params:
                batch_params["after_uid"] = name_to_uid(batch_params["after_uid"])

            uids = [name_to_uid(_) for _ in selection_seq]
            if success:
                items_moved, qsize_after = await pq.move_batch(uids=uids, **batch_params)
                items_seq = "".join([_["name"] for _ in items_moved])
                assert items_seq == batch_seq
                assert qsize_after == qsize
            else:
                with pytest.raises(Exception, match=msg):
                    await pq.move_batch(uids=uids, **batch_params)

            # Make sure that the queue is in the correct state
            queue, _ = await pq.get_queue()
            seq = "".join([_["name"] for _ in queue])
            assert seq == expected_seq
            assert len(queue) == qsize

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("pos, name", [
    ("front", "a"),
    ("back", "c"),
    (0, "a"),
    (1, "b"),
    (2, "c"),
    (3, None),  # Index out of range
    (-1, "c"),
    (-2, "b"),
    (-3, "a"),
    (-4, None)  # Index out of range
])
# fmt: on
def test_pop_item_from_queue_1(pos, name):
    """
    Basic test for the function ``PlanQueueOperations.pop_item_from_queue()``
    """

    async def testing():
        async with PQ() as pq:
            await pq.add_item_to_queue({"name": "a"})
            await pq.add_item_to_queue({"name": "b"})
            await pq.add_item_to_queue({"name": "c"})
            assert await pq.get_queue_size() == 3

            pq_uid = pq.plan_queue_uid

            if name is not None:
                plan, qsize = await pq.pop_item_from_queue(pos=pos)
                assert plan["name"] == name
                assert qsize == 2
                assert await pq.get_queue_size() == 2
                # Push the plan back to the queue (proves that UID is removed from '_uid_dict')
                await pq.add_item_to_queue(plan)
                assert await pq.get_queue_size() == 3
                assert pq.plan_queue_uid != pq_uid
            else:
                with pytest.raises(IndexError, match="Index .* is out of range"):
                    await pq.pop_item_from_queue(pos=pos)
                assert pq.plan_queue_uid == pq_uid

    asyncio.run(testing())


@pytest.mark.parametrize("pos", ["front", "back", 0, 1, -1])
def test_pop_item_from_queue_2(pos):
    """
    Test for the function ``PlanQueueOperations.pop_item_from_queue()``:
    the case of empty queue.
    """

    async def testing():
        async with PQ() as pq:
            assert await pq.get_queue_size() == 0
            pq_uid = pq.plan_queue_uid
            with pytest.raises(IndexError, match="Index .* is out of range|Queue is empty"):
                await pq.pop_item_from_queue(pos=pos)
            assert pq.plan_queue_uid == pq_uid

    asyncio.run(testing())


def test_pop_item_from_queue_3():
    """
    Pop plans by UID.
    """

    async def testing():
        async with PQ() as pq:
            await pq.add_item_to_queue({"item_type": "plan", "name": "a"})
            await pq.add_item_to_queue({"item_type": "plan", "name": "b"})
            await pq.add_item_to_queue({"item_type": "plan", "name": "c"})
            assert await pq.get_queue_size() == 3

            plans, _ = await pq.get_queue()
            assert len(plans) == 3
            plan_to_remove = [_ for _ in plans if _["name"] == "b"][0]

            # Remove one plan
            pq_uid = pq.plan_queue_uid
            await pq.pop_item_from_queue(uid=plan_to_remove["item_uid"])
            assert await pq.get_queue_size() == 2
            assert pq.plan_queue_uid != pq_uid

            # Attempt to remove the plan again. This should raise an exception.
            pq_uid = pq.plan_queue_uid
            with pytest.raises(
                IndexError, match=f"Plan with UID '{plan_to_remove['item_uid']}' " f"is not in the queue"
            ):
                await pq.pop_item_from_queue(uid=plan_to_remove["item_uid"])
            assert await pq.get_queue_size() == 2
            assert pq.plan_queue_uid == pq_uid

            # Attempt to remove the plan that is running. This should raise an exception.
            await pq.set_next_item_as_running()
            assert await pq.get_queue_size() == 1
            pq_uid = pq.plan_queue_uid
            with pytest.raises(IndexError, match="Can not remove an item which is currently running"):
                await pq.pop_item_from_queue(uid=plans[0]["item_uid"])
            assert await pq.get_queue_size() == 1
            assert pq.plan_queue_uid == pq_uid

    asyncio.run(testing())


def test_pop_item_from_queue_4_fail():
    """
    Failing tests for the function ``PlanQueueOperations.pop_item_from_queue()``
    """

    async def testing():
        async with PQ() as pq:
            pq_uid = pq.plan_queue_uid
            with pytest.raises(ValueError, match="Parameter 'pos' has incorrect value"):
                await pq.pop_item_from_queue(pos="something")
            assert pq.plan_queue_uid == pq_uid

            # Ambiguous parameters (position and UID is passed)
            with pytest.raises(ValueError, match="Ambiguous parameters"):
                await pq.pop_item_from_queue(pos=5, uid="abc")
            assert pq.plan_queue_uid == pq_uid

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("batch_params, queue_seq, selection_seq, batch_seq, expected_seq, success, msg", [
    ({}, "0123456", "", "", "0123456", True, ""),
    ({}, "0123456", "23", "23", "01456", True, ""),
    ({}, "0123456", "32", "32", "01456", True, ""),
    ({}, "0123456", "06", "06", "12345", True, ""),
    ({}, "0123456", "283", "23", "01456", True, ""),
    ({}, "0123456", "2893", "23", "01456", True, ""),
    ({}, "0123456", "2443", "243", "0156", True, ""),
    ({"ignore_missing": True}, "0123456", "2443", "243", "0156", True, ""),
    ({"ignore_missing": True}, "0123456", "283", "23", "01456", True, ""),
    ({"ignore_missing": False}, "0123456", "2443", "", "0123456", False, "The list of contains repeated UIDs"),
    ({"ignore_missing": False}, "0123456", "283", "", "0123456", False,
     "The queue does not contain items with the following UIDs"),
    ({"ignore_missing": False}, "0123456", "2883", "", "0123456", False, "The list of contains repeated UIDs"),
    ({}, "0123456", "", "", "0123456", True, ""),
    ({}, "", "", "", "", True, ""),
    ({}, "", "23", "", "", True, ""),
    ({"ignore_missing": False}, "", "", "", "", True, ""),
    ({"ignore_missing": False}, "", "23", "", "", False,
     "The queue does not contain items with the following UIDs"),
])
# fmt: on
def test_pop_items_from_queue_batch_1(
    batch_params, queue_seq, selection_seq, batch_seq, expected_seq, success, msg
):
    """
    Tests for ``pop_items_from_queue_batch``.
    """

    async def testing():
        async with PQ() as pq:

            async def add_plan(plan, n, **kwargs):
                plan_added, qsize = await pq.add_item_to_queue(plan, **kwargs)
                assert plan_added["name"] == plan["name"], f"plan: {plan}"
                assert qsize == n, f"plan: {plan}"

            def name_to_uid(uid):
                return f"{uid}{uid}"

            # Create the queue with plans
            for n, p_name in enumerate(queue_seq):
                await add_plan({"name": p_name, "item_uid": f"{name_to_uid(p_name)}"}, n + 1)
            qsize = await pq.get_queue_size()
            assert qsize == len(queue_seq)

            uids = [name_to_uid(_) for _ in selection_seq]
            if success:
                items_removed, qsize_after = await pq.pop_item_from_queue_batch(uids=uids, **batch_params)
                items_seq = "".join([_["name"] for _ in items_removed])
                assert items_seq == batch_seq
                assert qsize_after == len(expected_seq)
            else:
                with pytest.raises(Exception, match=msg):
                    await pq.pop_item_from_queue_batch(uids=uids, **batch_params)

            # Make sure that the queue is in the correct state
            queue, _ = await pq.get_queue()
            seq = "".join([_["name"] for _ in queue])
            assert seq == expected_seq
            assert len(queue) == len(expected_seq)

    asyncio.run(testing())


def test_clear_queue():
    """
    Test for ``PlanQueueOperations.clear_queue`` function
    """

    async def testing():
        async with PQ() as pq:
            await pq.add_item_to_queue({"item_type": "plan", "name": "a"})
            await pq.add_item_to_queue({"item_type": "plan", "name": "b"})
            await pq.add_item_to_queue({"item_type": "plan", "name": "c"})
            # Set one of 3 plans as running (removes it from the queue)
            await pq.set_next_item_as_running()

            assert await pq.get_queue_size() == 2
            assert len(pq._uid_dict) == 3

            pq_uid = pq.plan_queue_uid

            # Clears the queue only (doesn't touch the running plan)
            await pq.clear_queue()

            assert await pq.get_queue_size() == 0
            assert len(pq._uid_dict) == 1
            assert pq.plan_queue_uid != pq_uid

            with pytest.raises(ValueError, match="Parameter 'pos' has incorrect value"):
                await pq.pop_item_from_queue(pos="something")

    asyncio.run(testing())


def test_add_to_history_functions():
    """
    Test for ``PlanQueueOperations._add_to_history()`` method.
    """

    async def testing():
        async with PQ() as pq:
            assert await pq.get_history_size() == 0

            plans = [{"name": "a"}, {"name": "b"}, {"name": "c"}]
            ph_uid = pq.plan_history_uid
            for plan in plans:
                await pq._add_to_history(plan)
            assert await pq.get_history_size() == 3
            assert pq.plan_history_uid != ph_uid

            ph_uid = pq.plan_history_uid
            plan_history, plan_history_uid_1 = await pq.get_history()
            assert pq.plan_history_uid == plan_history_uid_1
            assert pq.plan_history_uid == ph_uid

            assert len(plan_history) == 3
            assert plan_history == plans

            ph_uid = pq.plan_history_uid
            await pq.clear_history()
            assert pq.plan_history_uid != ph_uid

            plan_history, _ = await pq.get_history()
            assert plan_history == []

    asyncio.run(testing())


@pytest.mark.parametrize("immediate_execution", [False, True])
@pytest.mark.parametrize("func", ["process_next_item", "set_next_item_as_running"])
@pytest.mark.parametrize("loop_mode", [False, True])
def test_process_next_item_1(func, loop_mode, immediate_execution):
    """
    Test for ``PlanQueueOperations.process_next_item()`` and
    ``PlanQueueOperations.set_next_item_as_running()`` functions.
    Those functions are interchangeable for plans, but ``process_next_item`` works
    for items and plans.
    """

    async def testing():
        async with PQ() as pq:
            await pq.set_plan_queue_mode({"loop": loop_mode})

            # Apply to empty queue
            assert await pq.get_queue_size() == 0
            assert await pq.is_item_running() is False
            assert await pq.set_next_item_as_running() == {}
            assert await pq.get_queue_size() == 0
            assert await pq.is_item_running() is False

            # Apply to a queue with several plans
            p1, _ = await pq.add_item_to_queue({"item_type": "plan", "name": "a"})
            p2, _ = await pq.add_item_to_queue({"item_type": "plan", "name": "b"})
            p3, _ = await pq.add_item_to_queue({"item_type": "plan", "name": "c"})

            # Plan used for immediate execution
            plan4 = {"item_type": "plan", "name": "d"}

            if func == "process_next_item":
                f = pq.process_next_item
            elif func == "set_next_item_as_running":
                f = pq.set_next_item_as_running
            else:
                assert False, f"Unknown test case: '{func}'"

            pq_uid = pq.plan_queue_uid
            queue_1 = await pq.get_queue()

            # If the next plan is set as running, it is removed from the queue and
            #   the queue UID is changed. If a plan is set for immediate execution,
            #   then the queue and the queue UID remain unchanged.
            params = {"item": plan4} if immediate_execution else {}
            running_plan = await f(**params)
            queue_2 = await pq.get_queue()

            assert running_plan != {}

            if immediate_execution:
                assert running_plan["name"] == "d"
                assert "properties" in running_plan
                assert running_plan["properties"]["immediate_execution"] is True
                assert pq.plan_queue_uid == pq_uid
                assert await pq.get_queue_size() == 3
                assert len(pq._uid_dict) == 3
                assert queue_1 == queue_2
            else:
                assert running_plan["name"] == "a"
                assert pq.plan_queue_uid != pq_uid
                assert await pq.get_queue_size() == 2
                assert len(pq._uid_dict) == 3

            # Apply if a plan is already running
            pq_uid = pq.plan_queue_uid
            assert await f() == {}
            assert pq.plan_queue_uid == pq_uid
            assert await pq.get_queue_size() == (3 if immediate_execution else 2)
            assert len(pq._uid_dict) == 3

    asyncio.run(testing())


@pytest.mark.parametrize("immediate_execution", [False, True])
@pytest.mark.parametrize("loop_mode", [False, True])
def test_process_next_item_2(loop_mode, immediate_execution):
    """
    Test for ``PlanQueueOperations.process_next_item()`` and
    ``PlanQueueOperations.set_next_item_as_running()`` functions.
    Those functions are interchangeable for plans, but ``process_next_item`` works
    for items and plans.

    Test with 'instruction' item.
    """

    async def testing():
        async with PQ() as pq:
            await pq.set_plan_queue_mode({"loop": loop_mode})

            # Apply to empty queue
            assert await pq.get_queue_size() == 0
            assert await pq.is_item_running() is False
            assert await pq.set_next_item_as_running() == {}
            assert await pq.get_queue_size() == 0
            assert await pq.is_item_running() is False

            # Apply to a queue with several plans
            p0, _ = await pq.add_item_to_queue({"item_type": "instruction", "name": "a"})
            p1, _ = await pq.add_item_to_queue({"item_type": "plan", "name": "b"})
            p2, _ = await pq.add_item_to_queue({"item_type": "plan", "name": "c"})

            item4 = {"item_type": "instruction", "name": "d"}

            pq_uid = pq.plan_queue_uid
            queue_1 = await pq.get_queue()

            params = {"item": item4} if immediate_execution else {}
            running_item = await pq.process_next_item(**params)
            queue_2 = await pq.get_queue()

            if immediate_execution:
                assert {_: running_item[_] for _ in ("name", "item_type")} == item4
                assert pq.plan_queue_uid == pq_uid
                assert await pq.get_queue_size() == 3
                assert len(pq._uid_dict) == 3
                assert queue_2 == queue_1
            else:
                assert running_item == p0
                assert pq.plan_queue_uid != pq_uid
                assert await pq.get_queue_size() == 3 if loop_mode else 2
                assert len(pq._uid_dict) == 3 if loop_mode else 2

                queue, _ = await pq.get_queue()
                assert queue[0] == p1
                assert queue[1] == p2
                if loop_mode:
                    assert queue[2]["name"] == p0["name"]
                    assert queue[2]["item_uid"] != p0["name"]

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("has_uid", [False, True])
@pytest.mark.parametrize("item_type", ["plan", "instruction"])
# fmt: on
def test_process_next_item_3(item_type, has_uid):
    """
    Verify if the function ``process_next_item`` is assigning new UID to plans, but not instructions.
    """

    async def testing():
        async with PQ() as pq:
            # Apply to a queue with several plans
            await pq.add_item_to_queue({"item_type": "instruction", "name": "a"})
            await pq.add_item_to_queue({"item_type": "plan", "name": "b"})
            await pq.add_item_to_queue({"item_type": "plan", "name": "c"})

            item4 = {"item_type": item_type, "name": "d"}
            if has_uid:
                item4 = pq.set_new_item_uuid(item4)

            running_item = await pq.process_next_item(item=item4)

            if has_uid:
                assert running_item["item_uid"] == item4["item_uid"]
            else:
                assert "item_uid" in running_item
                assert isinstance(running_item["item_uid"], str)

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("immediate_execution", [False, True])
# fmt: on
def test_process_next_item_4_fail(immediate_execution):
    """
    Try using 'set_next_item_as_running' with an instruction. Exception should be raised.
    The queue should remain unchanged.
    """

    async def testing():
        async with PQ() as pq:
            # Apply to a queue with several plans
            await pq.add_item_to_queue({"item_type": "instruction", "name": "a"})
            await pq.add_item_to_queue({"item_type": "plan", "name": "b"})
            await pq.add_item_to_queue({"item_type": "plan", "name": "c"})

            item4 = {"item_type": "instruction", "name": "d"}

            pq_uid = pq.plan_queue_uid
            queue_1, _ = await pq.get_queue()

            params = {"item": item4} if immediate_execution else {}

            with pytest.raises(RuntimeError, match="was called for an item other than plan"):
                await pq.set_next_item_as_running(**params)

            queue_2, _ = await pq.get_queue()  # The queue should remain untouched

            assert pq.plan_queue_uid == pq_uid
            assert await pq.get_queue_size() == 3
            assert len(pq._uid_dict) == 3
            assert queue_2 == queue_1

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("item_in, item_out", [
    ({"name": "count"}, {"name": "count"}),
    ({"name": "count", "properties": {}}, {"name": "count"}),
    ({"name": "count", "properties": {"nonexisting": 10}}, {"name": "count", "properties": {"nonexisting": 10}}),
    ({"name": "count", "properties": {"immediate_execution": True}}, {"name": "count"}),
])
# fmt: on
def test_clean_item_properties_1(item_in, item_out):
    """
    Basic test for `_clean_item_properties` function.
    """

    async def testing():
        async with PQ() as pq:
            item_cleaned = pq._clean_item_properties(item_in)
            assert item_cleaned == item_out

    asyncio.run(testing())


def test_set_processed_item_as_completed_1():
    """
    Test for ``PlanQueueOperations.set_processed_item_as_completed()`` function.
    The function moves currently running plan to history.
    """

    plans = [
        {"item_type": "plan", "item_uid": 1, "name": "a"},
        {"item_type": "plan", "item_uid": 2, "name": "b"},
        {"item_type": "plan", "item_uid": 3, "name": "c"},
    ]
    plans_run_uids = [["abc1"], ["abc2", "abc3"], []]

    def add_status_to_plans(plans, run_uids, exit_status):
        plans = copy.deepcopy(plans)
        plans_modified = []
        for plan, run_uid in zip(plans, run_uids):
            plan.setdefault("result", {})
            plan["result"]["exit_status"] = exit_status
            plan["result"]["run_uids"] = run_uid
            plans_modified.append(plan)
        return plans_modified

    async def testing():
        async with PQ() as pq:
            for plan in plans:
                await pq.add_item_to_queue(plan)

            # No plan is running
            pq_uid = pq.plan_queue_uid
            ph_uid = pq.plan_history_uid
            plan = await pq.set_processed_item_as_completed(exit_status="completed", run_uids=plans_run_uids[0])
            assert plan == {}
            assert pq.plan_queue_uid == pq_uid
            assert pq.plan_history_uid == ph_uid

            # Execute the first plan
            await pq.set_next_item_as_running()
            pq_uid = pq.plan_queue_uid
            plan = await pq.set_processed_item_as_completed(exit_status="completed", run_uids=plans_run_uids[0])
            assert pq.plan_queue_uid != pq_uid
            assert pq.plan_history_uid != ph_uid

            assert await pq.get_queue_size() == 2
            assert await pq.get_history_size() == 1
            assert plan["name"] == plans[0]["name"]
            assert plan["result"]["exit_status"] == "completed"
            assert plan["result"]["run_uids"] == plans_run_uids[0]

            plan_history, _ = await pq.get_history()
            plan_history_expected = add_status_to_plans(plans[0:1], plans_run_uids[0:1], "completed")
            assert plan_history == plan_history_expected

            # Execute the second plan
            await pq.set_next_item_as_running()
            plan = await pq.set_processed_item_as_completed(exit_status="completed", run_uids=plans_run_uids[1])

            assert await pq.get_queue_size() == 1
            assert await pq.get_history_size() == 2
            assert plan["name"] == plans[1]["name"]
            assert plan["result"]["exit_status"] == "completed"
            assert plan["result"]["run_uids"] == plans_run_uids[1]

            plan_history, _ = await pq.get_history()
            plan_history_expected = add_status_to_plans(plans[0:2], plans_run_uids[0:2], "completed")
            assert plan_history == plan_history_expected

    asyncio.run(testing())


def test_set_processed_item_as_completed_2():
    """
    Test for ``PlanQueueOperations.set_processed_item_as_completed()`` function.
    Similar test as the previous one, but with LOOP mode ENABLED.
    """

    plans = [
        {"item_type": "plan", "item_uid": 1, "name": "a"},
        {"item_type": "plan", "item_uid": 2, "name": "b"},
        {"item_type": "plan", "item_uid": 3, "name": "c"},
    ]
    plans_run_uids = [["abc1"], ["abc2", "abc3"], []]

    async def testing():
        async with PQ() as pq:
            for plan in plans:
                await pq.add_item_to_queue(plan)

            # Enable LOOP mode: the completed item will be pushed to the back of the queue
            await pq.set_plan_queue_mode({"loop": True})

            # No plan is running
            pq_uid = pq.plan_queue_uid
            ph_uid = pq.plan_history_uid
            plan = await pq.set_processed_item_as_completed(exit_status="completed", run_uids=plans_run_uids[0])
            assert plan == {}
            assert pq.plan_queue_uid == pq_uid
            assert pq.plan_history_uid == ph_uid

            queue, _ = await pq.get_queue()
            assert [_["name"] for _ in queue] == ["a", "b", "c"]

            # Execute the first plan
            await pq.set_next_item_as_running()
            pq_uid = pq.plan_queue_uid
            plan = await pq.set_processed_item_as_completed(exit_status="completed", run_uids=plans_run_uids[0])
            assert pq.plan_queue_uid != pq_uid
            assert pq.plan_history_uid != ph_uid

            queue, _ = await pq.get_queue()
            assert [_["name"] for _ in queue] == ["b", "c", "a"]

            assert await pq.get_queue_size() == 3
            assert await pq.get_history_size() == 1
            assert plan["name"] == plans[0]["name"]
            assert plan["result"]["exit_status"] == "completed"
            assert plan["result"]["run_uids"] == plans_run_uids[0]

            # Execute the second plan
            await pq.set_next_item_as_running()
            plan = await pq.set_processed_item_as_completed(exit_status="completed", run_uids=plans_run_uids[1])

            queue, _ = await pq.get_queue()
            assert [_["name"] for _ in queue] == ["c", "a", "b"]

            assert await pq.get_queue_size() == 3
            assert await pq.get_history_size() == 2
            assert plan["name"] == plans[1]["name"]
            assert plan["result"]["exit_status"] == "completed"
            assert plan["result"]["run_uids"] == plans_run_uids[1]

    asyncio.run(testing())


def test_set_processed_item_as_stopped_1():
    """
    Test for ``PlanQueueOperations.set_processed_item_as_stopped()`` function.
    The function pushes running plan back to the queue and saves it in history as well.

    Typically execution of single-run plans result in no UIDs, but in this test we still assign UIDS
    to test functionality.
    """
    plans = [
        {"item_type": "plan", "item_uid": "1", "name": "a"},
        {"item_type": "plan", "item_uid": "2", "name": "b"},
        {"item_type": "plan", "item_uid": "3", "name": "c"},
    ]
    plans_run_uids = [["abc1"], ["abc2", "abc3"], []]

    def add_status_to_plans(plans, run_uids, exit_status):
        plans = copy.deepcopy(plans)
        plans_modified = []
        for plan, run_uid in zip(plans, run_uids):
            plan.setdefault("result", {})
            plan["result"]["exit_status"] = exit_status
            plan["result"]["run_uids"] = run_uid
            plans_modified.append(plan)
        return plans_modified

    async def testing():
        async with PQ() as pq:
            for plan in plans:
                await pq.add_item_to_queue(plan)

            # No plan is running
            pq_uid = pq.plan_queue_uid
            ph_uid = pq.plan_history_uid
            plan = await pq.set_processed_item_as_stopped(exit_status="stopped", run_uids=plans_run_uids[0])
            assert plan == {}
            assert pq.plan_queue_uid == pq_uid
            assert pq.plan_history_uid == ph_uid

            # Execute the first plan
            await pq.set_next_item_as_running()
            pq_uid = pq.plan_queue_uid
            plan = await pq.set_processed_item_as_stopped(exit_status="stopped", run_uids=plans_run_uids[0])
            assert pq.plan_queue_uid != pq_uid
            assert pq.plan_history_uid != ph_uid

            assert await pq.get_queue_size() == 3
            assert await pq.get_history_size() == 1
            assert plan["name"] == plans[0]["name"]
            assert plan["result"]["exit_status"] == "stopped"
            assert plan["result"]["run_uids"] == plans_run_uids[0]
            assert plan["item_uid"] == plans[0]["item_uid"]

            # New plan UID is generated when the plan is pushed back into the queue
            queue, _ = await pq.get_queue()
            plan_modified_uid = queue[0]["item_uid"]
            # Make sure that the item UID was changed
            assert plan_modified_uid != plans[0]["item_uid"]

            plan_history, _ = await pq.get_history()
            plan_history_expected = add_status_to_plans([plans[0]], [plans_run_uids[0]], "stopped")
            assert plan_history == plan_history_expected

            # Execute the second plan
            await pq.set_next_item_as_running()
            plan = await pq.set_processed_item_as_stopped(exit_status="stopped", run_uids=plans_run_uids[1])

            assert await pq.get_queue_size() == 3
            assert await pq.get_history_size() == 2
            assert plan["name"] == plans[0]["name"]
            assert plan["result"]["exit_status"] == "stopped"
            assert plan["result"]["run_uids"] == plans_run_uids[1]

            plan_history, _ = await pq.get_history()
            plan_history_expected = add_status_to_plans(
                [plans[0].copy(), plans[0].copy()], [plans_run_uids[0], plans_run_uids[1]], "stopped"
            )
            # Plan 0 has different UID after it was inserted in the queue during the 1st attempt
            plan_history_expected[1]["item_uid"] = plan_modified_uid
            assert plan_history == plan_history_expected

            # Verify that `_uid_dict` still has correct size. `_uid_dict` should never be accessed directly.
            assert len(pq._uid_dict) == 3
            # Also it should not contain UIDs of already executed plans.
            for plan in plan_history:
                assert plan["item_uid"] not in pq._uid_dict

    asyncio.run(testing())


# fmt: off
@pytest.mark.parametrize("func", ["completed", "stopped"])
@pytest.mark.parametrize("loop_mode", [False, True])
# fmt: on
def test_set_processed_item_as_stopped_2(loop_mode, func):
    """
    ``set_processed_item_as_completed`` and ``set_processed_item_as_completed`` processing of an item set
    for immediate execution: basic functionality.
    """
    plans = [
        {"item_type": "plan", "item_uid": 1, "name": "a"},
        {"item_type": "plan", "item_uid": 2, "name": "b"},
        {"item_type": "plan", "item_uid": 3, "name": "c"},
    ]
    plan4 = {"item_type": "plan", "item_uid": 3, "name": "d"}
    plan4_run_uids = ["abc2", "abc3"]

    async def testing():
        async with PQ() as pq:
            for plan in plans:
                await pq.add_item_to_queue(plan)

            # Enable LOOP mode: the completed item will be pushed to the back of the queue
            await pq.set_plan_queue_mode({"loop": loop_mode})

            pq_uid = pq.plan_queue_uid
            ph_uid = pq.plan_history_uid
            queue_1, _ = await pq.get_queue()

            await pq.process_next_item(item=plan4)

            assert pq.plan_queue_uid == pq_uid
            assert pq.plan_history_uid == ph_uid

            if func == "completed":
                plan = await pq.set_processed_item_as_completed(exit_status=func, run_uids=plan4_run_uids)
            elif func == "stopped":
                plan = await pq.set_processed_item_as_stopped(exit_status=func, run_uids=plan4_run_uids)
            else:
                assert False, f"Unknown value of parameter 'func': '{func}'"

            queue_2, _ = await pq.get_queue()

            assert queue_1 == queue_2
            assert await pq.get_queue_size() == 3
            assert await pq.get_history_size() == 1
            assert pq.plan_queue_uid == pq_uid
            assert pq.plan_history_uid != ph_uid

            def check_plan(p):
                assert p["name"] == plan4["name"]
                assert p["result"]["exit_status"] == func
                assert p["result"]["run_uids"] == plan4_run_uids

            check_plan(plan)

            history, _ = await pq.get_history()
            check_plan(history[0])

            assert history[0]["item_uid"] == plan["item_uid"]

    asyncio.run(testing())


@pytest.mark.parametrize("loop_mode", [False, True])
@pytest.mark.parametrize("func", ["completed", "stopped"])
def test_set_processed_item_as_stopped_3(loop_mode, func):
    """
    Test for ``PlanQueueOperations.set_processed_item_as_completed()`` and
    ``PlanQueueOperations.set_processed_item_as_stopped()`` function.
    Make sure that ``item["properties"]["immediate_execution"] is removed before the plan
    is added to history and the plan is not pushed back into the queue if the plan is
    stopped. The behavior of both functions is identical for the plans executed
    in 'immediate execution' mode.
    """

    plan = {"item_type": "plan", "item_uid": 1, "name": "a", "properties": {"immediate_execution": True}}

    async def testing():
        async with PQ() as pq:
            nonlocal plan
            await pq.add_item_to_queue(plan)
            await pq.set_plan_queue_mode({"loop": loop_mode})

            assert await pq.get_queue_size() == 1
            assert await pq.get_history_size() == 0

            await pq.set_next_item_as_running()

            assert await pq.get_queue_size() == 0
            assert await pq.get_history_size() == 0

            if func == "completed":
                plan1 = await pq.set_processed_item_as_completed(exit_status="completed", run_uids=["abc"])
            elif func == "stopped":
                plan1 = await pq.set_processed_item_as_stopped(exit_status="stopped", run_uids=["abc"])
            else:
                raise Exception(f"Unsupported parameter value func={func!r}")

            # assert False, plan1
            assert plan1["name"] == plan["name"]
            assert "properties" not in plan1

            assert await pq.get_queue_size() == 0
            assert await pq.get_history_size() == 1

            history, _ = await pq.get_history()
            assert history[0] == plan1
            assert "properties" not in history[0]

    asyncio.run(testing())


# ==============================================================================================
#                    Saving and retrieving user group permissions
def test_user_group_permissions_1():
    """
    Test for saving and retrieving and clearing user group permissions, which are backed up in Redis.
    """
    ug_permissions_1 = {"some_key_1": "some_value_1"}
    ug_permissions_2 = {"some_key_2": "some_value_2"}

    async def testing():
        async with PQ() as pq:
            assert await pq.user_group_permissions_retrieve() is None

            await pq.user_group_permissions_save(ug_permissions_1)
            assert await pq.user_group_permissions_retrieve() == ug_permissions_1

            await pq.user_group_permissions_save(ug_permissions_2)
            assert await pq.user_group_permissions_retrieve() == ug_permissions_2

            await pq.user_group_permissions_clear()
            assert await pq.user_group_permissions_retrieve() is None

    asyncio.run(testing())



import copy
import numpy as np
import os
import pytest
import pprint

from bluesky_queueserver.manager.tests.plan_lists import plan_list_sample, create_excel_file_from_plan_list

from bluesky_queueserver.manager.conversions import (
    simplify_plan_descriptions,
    _read_cell_parameter,
    spreadsheet_to_plan_list,
)


# fmt: off
@pytest.mark.parametrize("plans_in, plans_out_expected", [
    # Two trivial plans (only name and no parameters). Checks if multiple plans are processed
    ({"plan1": {"name": "plan1"},
      "plan2": {"name": "plan1"}},

     {"plan1": {"name": "plan1"},
      "plan2": {"name": "plan1"}}),

    # Check recognition of 'int' scalar type
    ({"plan1": {"name": "plan1",
                "description": "Description of Plan 1.",
                "parameters": [
                    {"name": "p1",
                     "kind": "POSITIONAL_OR_KEYWORD",
                     "annotation": {"type": "int"}}
                ]}},
     {"plan1": {"name": "plan1",
                "description": "Description of Plan 1.",
                "parameters": [
                    {"name": "p1",
                     "kind": "POSITIONAL_OR_KEYWORD",
                     "type": "int",
                     "is_list": False,
                     "is_optional": False}
                ]}},
     ),

    # Check recognition of list of 'floats'
    ({"plan1": {"name": "plan1",
                "description": "Description of Plan 1.",
                "parameters": [
                    {"name": "p1",
                     "description": "Description for parameter 'p1'.",
                     "kind": "POSITIONAL_OR_KEYWORD",
                     "annotation": {"type": "typing.List[float]"},
                     "default": "[10, 20]",
                     "min": "5",
                     "max": "50",
                     "step": "0.01"
                     },
                    {"name": "p2",
                     "description": "Description for parameter 'p2'.",
                     "kind": "POSITIONAL_OR_KEYWORD",
                     "annotation": {"type": "List[float]"},
                     "default": "[30, 40]",
                     },
                    {"name": "p3",
                     "description": "Description for parameter 'p3'.",
                     "kind": "POSITIONAL_OR_KEYWORD",
                     "default": "10",
                     },
                ]},
      },
     {"plan1": {"name": "plan1",
                "description": "Description of Plan 1.",
                "parameters": [
                    {"name": "p1",
                     "description": "Description for parameter 'p1'.",
                     "kind": "POSITIONAL_OR_KEYWORD",
                     "type": "float",
                     "default": [10, 20],
                     "is_list": True,
                     "is_optional": True,
                     "min": 5,
                     "max": 50,
                     "step": 0.01},
                    {"name": "p2",
                     "description": "Description for parameter 'p2'.",
                     "kind": "POSITIONAL_OR_KEYWORD",
                     "type": "float",
                     "default": [30, 40],
                     "is_list": True,
                     "is_optional": True},
                    {"name": "p3",
                     "description": "Description for parameter 'p3'.",
                     "kind": "POSITIONAL_OR_KEYWORD",
                     "default": 10,
                     "is_optional": True},
                ]},
      },
     ),

    # Check enum
    ({"plan1": {"name": "plan1",
                "description": "Description of Plan 1.",
                "parameters": [
                    {"name": "p1",
                     "kind": "POSITIONAL_OR_KEYWORD",
                     "description": "Description for parameter 'p1'.",
                     "annotation": {
                         "type": "typing.List[typing.Union[Motor1, Motor2, Plan, Enum1, Enum2]]",
                         "devices": {"Motor1": ("m1", "m2"), "Motor2": ("m3")},
                         "plans": {"Plan": ("p1", "p2")},
                         "enums": {"Enum1": ("e1"), "Enum2": ("e2", "e3")},
                     }}
                ]}},
     {"plan1": {"name": "plan1",
                "description": "Description of Plan 1.",
                "parameters": [
                    {"name": "p1",
                     "description": "Description for parameter 'p1'.",
                     "kind": "POSITIONAL_OR_KEYWORD",
                     "type": "str",
                     "enum": ["m1", "m2", "m3", "p1", "p2", "e1", "e2", "e3"],
                     "is_list": True,
                     "is_optional": False}
                ]}},
     ),
])
# fmt: on
def test_simplify_plan_descriptions_1(plans_in, plans_out_expected):
    """
    Function ``filter_plan_descriptions``. Basic test.
    """
    plans_out = simplify_plan_descriptions(plans_in)
    assert plans_out == plans_out_expected, pprint.pformat(plans_out)


# fmt: off
@pytest.mark.parametrize("val_in, val_out, val_type, success, errmsg", [
    (10, 10, int, True, ""),
    (np.int64(10), 10, int, True, ""),
    (10.0, 10, int, True, ""),  # The number is expected to be rounded to int and represented as int
    (np.float64(10.0), 10, int, True, ""),
    (10.5, 10.5, float, True, ""),
    (np.float64(10.5), 10.5, float, True, ""),
    ("10", 10, int, True, ""),  # Number representation is more straightforward if they are strings
    ("10.0", 10, float, True, ""),
    ("10.5", 10.5, float, True, ""),
    ("10, 20", (10, 20), None, True, ""),
    ("10.3, 20", (10.3, 20), None, True, ""),
    ("'det1'", 'det1', str, True, ""),
    ("'det1', 'det2'", ('det1', 'det2'), None, True, ""),
    ("det1", 'det1', None, False, "name 'det1' is not defined"),  # detector names must be quoted
    ("det1, det2", ('det1', 'det2'), None, False, "name 'det1' is not defined"),  # detector names must be quoted
    ([10, 20], None, None, False, "Cell value .* has unsupported type"),
    ("[10, 20, 'det']", [10, 20, "det"], list, True, ""),  # List represented as a string
    # Test for a complicated expression
    ("{'a':'10','b':'20.5','c':'det','d':{'e':[50, 60]}}",
     {'a': '10', 'b': '20.5', 'c': 'det', 'd': {'e': [50, 60]}}, dict, True, ""),

])
# fmt: on
def test_read_cell_parameter(val_in, val_out, val_type, success, errmsg):
    if success:
        val_result = _read_cell_parameter(val_in)
        assert val_result == val_out
        if val_type:
            assert type(val_result) == val_type
    else:
        with pytest.raises(Exception, match=errmsg):
            _read_cell_parameter(val_in)


# fmt: on
@pytest.mark.parametrize("ext", [".xlsx", ".csv"])
# fmt: off
def test_spreadsheet_to_plan_list_1(tmp_path, ext):
    """
    Basic test for ``spreadsheet_to_plan_list()`` function. Check that the list of
    plans could be saved and then restored from the spreadsheet without change.
    Check that the restored parameters have correct types.
    """
    # Add plans with a kwarg that accepts values of different types (int, float and str)
    #   Those values are going to be placed in the same column of the spreadsheet.
    #   Check if types will be restored correctly (pandas write/read functions don't like columns
    #   with mixed types). Also check if all possible types are handled correctly.
    extra_plans = [
        {"name": "count", "args": [["det2"]], "kwargs": {"num": 2, "extra_param": 50}, "item_type": "plan"},
        {"name": "count", "args": [["det2"]], "kwargs": {
            "num": 2, "extra_param": "some_str"}, "item_type": "plan"},
        {"name": "count", "args": [["det2"]], "kwargs": {"num": 2, "extra_param": 50.256}, "item_type": "plan"},
        {"name": "count", "args": [["det2"]], "kwargs": {"num": 2, "extra_param": None}, "item_type": "plan"},
        {"name": "count", "args": [["det2"]], "kwargs": {"num": 2, "extra_param": ""}, "item_type": "plan"},
        {"name": "count", "args": [["det2"]], "kwargs": {
            "num": 2, "extra_param": [10, 20, 30]}, "item_type": "plan"},
        {"name": "count", "args": [["det2"]], "kwargs": {"num": 2, "extra_param": {
            "p1": 10, "p2": "10", "p3": 50}}, "item_type": "plan"},
    ]

    plan_list = copy.deepcopy(plan_list_sample)  # We are going to change the plans
    for plan in extra_plans:
        plan_list.append(plan)

    for plan in plan_list:
        if isinstance(plan["name"], str):
            plan["item_type"] = "plan"

    ss_path = create_excel_file_from_plan_list(tmp_path, plan_list=plan_list, ss_ext=ext)
    with open(ss_path, "rb") as f:
        plans = spreadsheet_to_plan_list(spreadsheet_file=f, file_name=os.path.split(ss_path)[-1])

    # Remove the case when name == None before comparing dictionaries
    assert plans == [_ for _ in plan_list if isinstance(_["name"], str)]


# fmt: on
@pytest.mark.parametrize("ext", [".xlsx", ".csv"])
@pytest.mark.parametrize(
    "extra_plan, errmsg",
    [
        # Invalid plan name (wrong type - not str)
        (
            {"name": 10, "args": [["det1", "det2"]], "kwargs": {"num": 10}},
            "Plan name .*row 9.* has incorrect type",
        ),
        (
            {"name": [10, 20], "args": [["det1", "det2"]], "kwargs": {"num": 10}},
            "Plan name .*row 9.* has incorrect type",
        ),
        # Invalid plan name (can not be a Python identifier)
        (
            {"name": "10", "args": [["det1", "det2"]], "kwargs": {"num": 10}},
            "Plan name .*10.*row 9.* is not a valid plan name",
        ),
        (
            {"name": "co unt", "args": [["det1", "det2"]], "kwargs": {"num": 10}},
            "Plan name .*co unt.*row 9.* is not a valid plan name",
        ),
    ],
)
# fmt: off
def test_spreadsheet_to_plan_list_2_fail(tmp_path, ext, extra_plan, errmsg):
    """
    Test if ``spreadsheet_to_plan_list()`` function is correctly processing the main issues
    with the spreadsheet and raises exception with correct error messages.
    """
    _plan_list_modified = plan_list_sample.copy()
    _plan_list_modified.append(extra_plan)
    ss_path = create_excel_file_from_plan_list(tmp_path, plan_list=_plan_list_modified, ss_ext=ext)
    with open(ss_path, "rb") as f:
        with pytest.raises(Exception, match=errmsg):
            spreadsheet_to_plan_list(spreadsheet_file=f, file_name=os.path.split(ss_path)[-1])



import pytest
import uuid

from bluesky_queueserver.manager.plan_monitoring import RunList, CallbackRegisterRun


def test_RunList_1():
    """
    Full functionality test: ``RunList`` class.
    """
    uids = [str(uuid.uuid4()) for _ in range(3)]
    is_open = [True] * 3
    exit_code = [None] * 3
    expected_run_list = [
        {"uid": _[0], "is_open": _[1], "exit_status": _[2]} for _ in zip(uids, is_open, exit_code)
    ]

    # Create run list
    run_list = RunList()
    assert run_list.is_changed() is False
    assert run_list.get_run_list() == []

    # Add one object
    run_list.add_run(uid=uids[0])
    assert run_list.is_changed() is True
    assert run_list.get_run_list() == expected_run_list[0:1]
    assert run_list.is_changed() is True
    assert run_list.get_run_list(clear_state=True) == expected_run_list[0:1]
    assert run_list.is_changed() is False

    # Add two more objects
    run_list.add_run(uid=uids[1])
    run_list.add_run(uid=uids[2])
    assert run_list.is_changed() is True
    assert run_list.get_run_list(clear_state=True) == expected_run_list
    assert run_list.is_changed() is False

    # Set the second object as completed
    expected_run_list[1]["is_open"] = False
    expected_run_list[1]["exit_status"] = "success"
    run_list.set_run_closed(uid=uids[1], exit_status="success")
    assert run_list.is_changed() is True
    assert run_list.get_run_list(clear_state=True) == expected_run_list
    assert run_list.is_changed() is False

    # Fail case: non-existing UID
    with pytest.raises(Exception, match="Run with UID .* was not found in the list"):
        run_list.set_run_closed(uid="non-existing-uid", exit_status="success")
    assert run_list.is_changed() is False

    run_list.clear()
    assert run_list.is_changed() is True
    assert run_list.get_run_list(clear_state=True) == []
    assert run_list.is_changed() is False


def test_CallbackRegisterRun_1():
    """
    Basic test: ``CallbackRegisterRun`` class.
    """
    run_list = RunList()
    cb = CallbackRegisterRun(run_list=run_list)
    uid = str(uuid.uuid4())

    cb("start", {"uid": uid})
    assert run_list.get_run_list() == [{"uid": uid, "is_open": True, "exit_status": None}]
    cb("stop", {"run_start": uid, "exit_status": "success"})
    assert run_list.get_run_list() == [{"uid": uid, "is_open": False, "exit_status": "success"}]



import argparse
from multiprocessing import Pipe, Queue
import threading
import time as ttime
import os
from importlib.util import find_spec

from .worker import RunEngineWorker
from .manager import RunEngineManager
from .comms import PipeJsonRpcReceive, validate_zmq_key
from .profile_ops import get_default_startup_dir
from .output_streaming import PublishConsoleOutput, setup_console_output_redirection
from .logging_setup import setup_loggers

from .. import __version__

import logging

logger = logging.getLogger(__name__)


class WatchdogProcess:
    def __init__(
        self,
        *,
        config_worker=None,
        config_manager=None,
        cls_run_engine_worker=RunEngineWorker,
        cls_run_engine_manager=RunEngineManager,
        msg_queue=None,
        log_level=logging.DEBUG,
    ):

        self._log_level = log_level

        self._cls_run_engine_worker = cls_run_engine_worker
        self._cls_run_engine_manager = cls_run_engine_manager

        self._msg_queue = msg_queue

        self._re_manager = None
        self._re_worker = None

        # The number of restarts of the manager processes including the first start.
        self._re_manager_n_restarts = 0

        # Create pipes used for connections of the modules
        self._manager_conn = None  # Worker -> Manager
        self._worker_conn = None  # Manager -> Worker
        self._watchdog_to_manager_conn = None  # Watchdog -> Manager
        self._manager_to_watchdog_conn = None  # Manager -> Watchdog
        self._create_conn_pipes()

        # Class that supports communication over the pipe
        self._comm_to_manager = PipeJsonRpcReceive(
            conn=self._watchdog_to_manager_conn, name="RE Watchdog-Manager Comm"
        )

        self._watchdog_state = 0  # State is currently just time since last notification
        self._watchdog_state_lock = threading.Lock()

        self._manager_is_stopping = False  # Set True to stop the server
        self._heartbeat_timeout = 5  # Time to wait before restarting RE Manager

        # Configuration of the RE environment (passed to RE Worker)
        self._config_worker = config_worker
        self._config_manager = config_manager

    def _create_conn_pipes(self):
        # Manager to worker
        self._manager_conn, self._worker_conn = Pipe()
        # Watchdog to manager
        self._watchdog_to_manager_conn, self._manager_to_watchdog_conn = Pipe()

    # ======================================================================
    #             Handlers for messages from RE Manager

    def _start_re_worker_handler(self, user_group_permissions):
        """
        Creates worker process. This is a quick operation, because it starts RE Worker
        process without waiting for initialization.
        """
        logger.info("Starting RE Worker ...")
        try:
            self._re_worker = self._cls_run_engine_worker(
                conn=self._manager_conn,
                name="RE Worker Process",
                config=self._config_worker,
                msg_queue=self._msg_queue,
                log_level=self._log_level,
                user_group_permissions=user_group_permissions,
            )
            self._re_worker.start()
            success, err_msg = True, ""
        except Exception as ex:
            success, err_msg = False, str(ex)
        return {"success": success, "err_msg": err_msg}

    def _join_re_worker_handler(self, *, timeout=0.5):
        """
        Join RE Worker process after it was orderly closed by RE Manager. Watchdog module doesn't
        communicate with the worker process directly. This is responsibility of the RE Manager.
        But RE Manager doesn't hold reference to RE Worker, so it needs to be joined here.
        """
        logger.info("Joining RE Worker ...")
        self._re_worker.join(timeout)  # Try to join with timeout
        success = not self._re_worker.is_alive()  # Return success
        return {"success": success}

    def _kill_re_worker_handler(self):
        """
        Kill RE Worker by request from RE Manager. This is done only if RE Worker is non-responsive
        and can not be orderly stopped.
        """
        # TODO: kill() or terminate()???
        logger.info("Killing RE Worker ...")
        self._re_worker.kill()
        self._re_worker.join()  # Not really necessary, but helps with unit testing.
        return {"success": True}

    def _is_worker_alive_handler(self):
        """
        Checks if RE Worker process is in running state. It doesn't mean that it is responsive.
        """
        is_alive = False
        if hasattr(self._re_worker, "is_alive"):
            is_alive = self._re_worker.is_alive()
        return {"worker_alive": is_alive}

    def _manager_stopping_handler(self):
        """
        Manager informed that it is stopping and should not be restarted.
        """
        self._manager_is_stopping = True

    def _init_watchdog_state(self):
        with self._watchdog_state_lock:
            self._watchdog_state = ttime.time()

    def _register_heartbeat_handler(self, *, value):
        """
        Heartbeat is received. Update the state.
        """
        if value == "alive":
            self._init_watchdog_state()

    # ======================================================================

    def _start_re_manager(self):
        self._re_manager_n_restarts += 1
        self._init_watchdog_state()
        self._re_manager = self._cls_run_engine_manager(
            conn_watchdog=self._manager_to_watchdog_conn,
            conn_worker=self._worker_conn,
            config=self._config_manager,
            name="RE Manager Process",
            msg_queue=self._msg_queue,
            log_level=self._log_level,
            number_of_restarts=self._re_manager_n_restarts,
        )
        self._re_manager.start()

    def run(self):

        logging.basicConfig(level=max(logging.WARNING, self._log_level))
        setup_loggers(log_level=self._log_level)

        # Requests
        self._comm_to_manager.add_method(self._start_re_worker_handler, "start_re_worker")
        self._comm_to_manager.add_method(self._join_re_worker_handler, "join_re_worker")
        self._comm_to_manager.add_method(self._kill_re_worker_handler, "kill_re_worker")
        self._comm_to_manager.add_method(self._is_worker_alive_handler, "is_worker_alive")
        # Notifications
        self._comm_to_manager.add_method(self._manager_stopping_handler, "manager_stopping")
        self._comm_to_manager.add_method(self._register_heartbeat_handler, "heartbeat")

        self._comm_to_manager.start()

        self._start_re_manager()
        while True:
            # Primitive implementation of the loop that restarts the process.
            self._re_manager.join(0.1)  # Small timeout

            if self._manager_is_stopping and not self._re_manager.is_alive():
                break  # Exit if the program was actually stopped (process joined)

            with self._watchdog_state_lock:
                time_passed = ttime.time() - self._watchdog_state

            # Interval is used to protect the system from restarting in case of clock issues.
            # It may be a better idea to implement a ticker in a separate thread to act as
            #   a clock to be completely independent from system clock.
            t_min, t_max = self._heartbeat_timeout, self._heartbeat_timeout + 10.0
            if (time_passed >= t_min) and (time_passed <= t_max) and not self._manager_is_stopping:
                logger.error("Timeout detected by Watchdog. RE Manager malfunctioned and must be restarted")
                self._re_manager.kill()
                self._start_re_manager()

        self._comm_to_manager.stop()
        logger.info("RE Watchdog is stopped")


def start_manager():

    s_enc = (
        "Encryption for ZeroMQ communication server may be enabled by setting QSERVER_ZMQ_PRIVATE_KEY\n"
        "environment variable to a valid private key (z85-encoded 40 character string):\n\n"
        "    export QSERVER_ZMQ_PRIVATE_KEY='<private_key>'\n\n"
        "A public/private key pair may be generated by running 'qserver-zmq-keys'. If RE Manager is\n"
        "configured to use encrypted channel, the encryption must also be enabled at the client side\n"
        "using the public key from the generated pair. Encryption is disabled by default."
    )

    def formatter(prog):
        # Set maximum width such that printed help mostly fits in the RTD theme code block (documentation).
        return argparse.RawDescriptionHelpFormatter(prog, max_help_position=20, width=90)

    parser = argparse.ArgumentParser(
        description=f"Start Run Engine (RE) Manager\nbluesky-queueserver version {__version__}\n\n{s_enc}",
        formatter_class=formatter,
    )
    parser.add_argument(
        "--zmq-addr",
        dest="zmq_addr",
        type=str,
        default="tcp://*:60615",
        help="The address of ZMQ server (control connection), e.g. 'tcp://*:60615' (default: %(default)s).",
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--startup-dir",
        dest="startup_dir",
        type=str,
        help="Path to directory that contains a set of startup files (*.py and *.ipy). All the scripts "
        "in the directory will be sorted in alphabetical order of their names and loaded in "
        "the Run Engine Worker environment. The set of startup files may be located in any accessible "
        "directory.",
    )
    group.add_argument(
        "--startup-profile",
        dest="profile_name",
        type=str,
        help="The name of IPython profile used to find the location of startup files. Example: if IPython is "
        "configured to look for profiles in '~/.ipython' directory (default behavior) and the profile "
        "name is 'testing', then RE Manager will look for startup files in "
        "'~/.ipython/profile_testing/startup' directory.",
    )
    group.add_argument(
        "--startup-module",
        dest="startup_module_name",
        type=str,
        help="The name of the module with startup code. The module is imported each time the RE Worker "
        "environment is opened. Example: 'some.startup.module'. Paths to the list of existing "
        "plans and devices (--existing-plans-and-devices) and user group permissions "
        "(--user-group-permissions) must be explicitly specified if this option is used.",
    )

    group.add_argument(
        "--startup-script",
        dest="startup_script_path",
        type=str,
        help="The path to the script with startup code. The script is loaded each time the RE Worker "
        "environment is opened. Example: '~/startup/scripts/scripts.py'. Paths to the list of existing "
        "plans and devices (--existing-plans-and-devices) and user group permissions "
        "(--user-group-permissions) must be explicitly specified if this option is used.",
    )

    parser.add_argument(
        "--existing-plans-devices",
        dest="existing_plans_and_devices_path",
        type=str,
        help="Path to file that contains the list of existing plans and devices. "
        "The path may be a relative path to the profile collection directory. "
        "If the path is directory, then the default file name "
        "'existing_plans_and_devices.yaml' is used.",
    )

    parser.add_argument(
        "--update-existing-plans-devices",
        dest="update_existing_plans_devices",
        type=str,
        choices=["NEVER", "ENVIRONMENT_OPEN", "ALWAYS"],
        default="ENVIRONMENT_OPEN",
        help="Select when the list of existing plans and devices stored on disk should be "
        "updated. The available choices are not to update the stored lists (NEVER), update "
        "the lists when the environment is opened (ENVIRONMENT_OPEN) or update the lists each "
        "the lists are changed (ALWAYS) "
        "(default: %(default)s)",
    )

    parser.add_argument(
        "--user-group-permissions",
        dest="user_group_permissions_path",
        type=str,
        help="Path to file that contains lists of plans and devices available to users. "
        "The path may be a relative path to the profile collection directory. "
        "If the path is a directory, then the default file name "
        "'user_group_permissions.yaml' is used.",
    )

    parser.add_argument(
        "--user-group-permissions-reload",
        dest="user_group_permissions_reload",
        type=str,
        choices=["NEVER", "ON_REQUEST", "ON_STARTUP"],
        default="ON_STARTUP",
        help="Select when user group permissions are reloaded from disk. Options: 'NEVER' - "
        "RE Manager never attempts to load permissions from disk file. If permissions fail to load "
        "from Redis, they are loaded from disk at the first startup of RE Manager or on request. "
        "'ON_REQUEST' - permissions are loaded from disk file when requested by 'permission_reload' API call. "
        "'ON_STARTUP' - permissions are loaded from disk each time RE Manager is started or when "
        "'permission_reload' API request is received "
        "(default: %(default)s)",
    )

    parser.add_argument(
        "--redis-addr",
        dest="redis_addr",
        type=str,
        default="localhost",
        help="The address of Redis server, e.g. 'localhost', '127.0.0.1', 'localhost:6379' "
        "(default: %(default)s). ",
    )

    parser.add_argument("--kafka-topic", dest="kafka_topic", type=str, help="The kafka topic to publish to.")
    parser.add_argument(
        "--kafka-server",
        dest="kafka_server",
        type=str,
        help="Bootstrap server to connect (default: %(default)s).",
        default="127.0.0.1:9092",
    )

    parser.add_argument(
        "--zmq-data-proxy-addr",
        dest="zmq_data_proxy_addr",
        type=str,
        help="The address of ZMQ proxy used to publish data. If the parameter is specified, RE is "
        "subscribed to 'bluesky.callbacks.zmq.Publisher' and documents are published via 0MQ proxy. "
        "0MQ Proxy (see Bluesky 0MQ documentation) should be started before plans are executed. "
        "The address should be in the form '127.0.0.1:5567' or 'localhost:5567'. The address is passed "
        "to 'bluesky.callbacks.zmq.Publisher'. It is recommended to use Kafka instead of 0MQ proxy in "
        "production data acquisition systems and use Kafka instead.",
    )

    parser.add_argument(
        "--keep-re",
        dest="keep_re",
        action="store_true",
        help="Keep RE created in profile collection. If the flag is set, RE must be "
        "created in the profile collection for the plans to run. RE will also "
        "keep all its subscriptions. Also must be subscribed to the Data Broker "
        "inside the profile collection, since '--databroker-config' argument "
        "is ignored.",
    )
    parser.add_argument(
        "--use-persistent-metadata",
        dest="use_persistent_metadata",
        action="store_true",
        help="Use msgpack-based persistent storage for scan metadata. Currently this "
        "is the preferred method to keep continuously incremented sequence of "
        "Run IDs between restarts of RE.",
    )
    parser.add_argument(
        "--databroker-config",
        dest="databroker_config",
        type=str,
        help="Name of the Data Broker configuration file.",
    )

    group_console_output = parser.add_argument_group(
        "Configure console output",
        "The arguments allow to configure printing and publishing of the console output\n"
        "generated by RE Manager. The arguments allow to set the address of 0MQ socket\n"
        "and enable/disable printing and/or publishing of the console output.",
    )

    group_console_output.add_argument(
        "--zmq-publish-console-addr",
        dest="zmq_publish_console_addr",
        type=str,
        default="tcp://*:60625",
        help="The address of ZMQ server (stdout and stderr publishing), e.g. 'tcp://*:60625' "
        "(default: %(default)s).",
    )

    group_console_output.add_argument(
        "--zmq-publish-console",
        dest="zmq_publish_console",
        type=str,
        choices=["ON", "OFF"],
        default="OFF",
        help="Enable (ON) or disable (OFF) publishing of console output to 0MQ (default: %(default)s).",
    )

    group_console_output.add_argument(
        "--console-output",
        dest="console_output",
        type=str,
        choices=["ON", "OFF"],
        default="ON",
        help="Enable (ON) or disable (OFF) printing of console output in the Re Manager terminal. "
        "(default: %(default)s)",
    )

    group_verbosity = parser.add_argument_group(
        "Logging verbosity settings",
        "The default logging settings (loglevel=INFO) provide optimal amount of data to monitor\n"
        "the operation of RE Manager. Select '--verbose' option to see detailed data on received and\n"
        "sent messages, added and executed plans, etc. Use options '--quiet' and '--silent'\n"
        "to see only warnings and error messages or disable logging output.",
    )
    group_v = group_verbosity.add_mutually_exclusive_group()
    group_v.add_argument(
        "--verbose",
        dest="logger_verbose",
        action="store_true",
        help="Set logger level to DEBUG.",
    )
    group_v.add_argument(
        "--quiet",
        dest="logger_quiet",
        action="store_true",
        help="Set logger level to WARNING.",
    )
    group_v.add_argument(
        "--silent",
        dest="logger_silent",
        action="store_true",
        help="Disables logging output.",
    )

    args = parser.parse_args()

    log_level = logging.INFO
    if args.logger_verbose:
        log_level = logging.DEBUG
    elif args.logger_quiet:
        log_level = logging.WARNING
    elif args.logger_silent:
        log_level = logging.CRITICAL + 1

    console_output_on = args.console_output == "ON"
    zmq_publish_console_on = args.zmq_publish_console == "ON"
    zmq_publish_console_addr = args.zmq_publish_console_addr

    msg_queue = Queue()
    setup_console_output_redirection(msg_queue)

    logging.basicConfig(level=max(logging.WARNING, log_level))
    setup_loggers(log_level=log_level)

    stream_publisher = PublishConsoleOutput(
        msg_queue=msg_queue,
        console_output_on=console_output_on,
        zmq_publish_on=zmq_publish_console_on,
        zmq_publish_addr=zmq_publish_console_addr,
    )

    if zmq_publish_console_on:
        # Wait for a short period to allow monitoring applications to connect.
        ttime.sleep(1)

    stream_publisher.start()

    config_worker = {}
    config_manager = {}
    if args.kafka_topic is not None:
        config_worker["kafka"] = {}
        config_worker["kafka"]["topic"] = args.kafka_topic
        config_worker["kafka"]["bootstrap"] = args.kafka_server

    if args.zmq_data_proxy_addr is not None:
        config_worker["zmq_data_proxy_addr"] = args.zmq_data_proxy_addr

    startup_dir, startup_module_name, startup_script_path = None, None, None

    # Find startup directory
    if args.profile_name:
        profile_name = args.profile_name
        if find_spec("IPython"):
            import IPython

            path_to_ipython = IPython.paths.get_ipython_dir()
        else:
            logger.error(
                "IPython is not installed. Specify directory to startup file by using '--startup-dir' option."
            )
            return 1
        ipython_dir = os.path.abspath(path_to_ipython)
        profile_name_full = f"profile_{profile_name}"
        startup_dir = os.path.join(ipython_dir, profile_name_full, "startup")
    elif args.startup_dir:
        startup_dir = args.startup_dir
        startup_dir = os.path.abspath(os.path.expanduser(startup_dir))
    elif args.startup_module_name:
        startup_module_name = args.startup_module_name
    elif args.startup_script_path:
        startup_script_path = os.path.abspath(os.path.expanduser(args.startup_script_path))
    else:
        # The default collection is the collection of simulated Ophyd devices
        #   and built-in Bluesky plans.
        startup_dir = get_default_startup_dir()

    if sum([_ is not None for _ in [startup_dir, startup_module_name, startup_script_path]]) != 1:
        logger.error("Multiple or no startup code sources were specified.")
        return 1

    # Primitive error processing: make sure that all essential data exists.
    if startup_dir is not None:
        if not os.path.exists(startup_dir):
            logger.error("Startup directory '%s' does not exist", startup_dir)
            return 1
        if not os.path.isdir(startup_dir):
            logger.error("Startup directory '%s' is not a directory", startup_dir)
            return 1
    elif (startup_module_name is not None) or (startup_script_path is not None):
        # startup_module_name or startup_script_path is set. This option requires
        #   the paths to existing plans and devices and user group permissions to be set.
        #   (The default directory can not be used in this case).
        if not args.existing_plans_and_devices_path:
            logger.error(
                "The path to the list of existing plans and devices (--existing-plans-and-devices) "
                "is not specified."
            )
            return 1
        if not args.user_group_permissions_path:
            logger.error(
                "The path to the file containing user group permissions (--user-group-permissions) "
                "is not specified."
            )
            return 1
        # Check if startup script exists (if it is specified)
        if startup_script_path is not None:
            if not os.path.isfile(startup_script_path):
                logger.error(f"The script '{startup_script_path}' is not found.")
                return 1

    config_worker["keep_re"] = args.keep_re
    config_worker["use_persistent_metadata"] = args.use_persistent_metadata

    config_worker["databroker"] = {}
    if args.databroker_config:
        config_worker["databroker"]["config"] = args.databroker_config

    config_worker["startup_dir"] = startup_dir
    config_worker["startup_module_name"] = startup_module_name
    config_worker["startup_script_path"] = startup_script_path

    default_existing_pd_fln = "existing_plans_and_devices.yaml"
    if args.existing_plans_and_devices_path:
        existing_pd_path = os.path.expanduser(args.existing_plans_and_devices_path)
        if not os.path.isabs(existing_pd_path) and startup_dir:
            existing_pd_path = os.path.join(startup_dir, existing_pd_path)
        if not existing_pd_path.endswith(".yaml"):
            existing_pd_path = os.path.join(existing_pd_path, default_existing_pd_fln)
    else:
        existing_pd_path = os.path.join(startup_dir, default_existing_pd_fln)
    # The file may not exist, but the directory MUST exist
    pd_dir = os.path.dirname(existing_pd_path) or "."
    if not os.path.isdir(os.path.dirname(existing_pd_path)):
        logger.error(
            "The directory for list of plans and devices ('%s')does not exist. "
            "Create the directory manually and restart RE Manager.",
            pd_dir,
        )
        return 1
    if not os.path.isfile(existing_pd_path):
        logger.warning(
            "The file with the list of allowed plans and devices ('%s') does not exist. "
            "The manager will be started with empty list. The list will be populated after "
            "RE worker environment is opened the first time.",
            existing_pd_path,
        )

    default_user_group_pd_fln = "user_group_permissions.yaml"
    if args.user_group_permissions_path:
        user_group_pd_path = os.path.expanduser(args.user_group_permissions_path)
        if not os.path.isabs(user_group_pd_path) and startup_dir:
            user_group_pd_path = os.path.join(startup_dir, user_group_pd_path)
        if not user_group_pd_path.endswith(".yaml"):
            user_group_pd_path = os.path.join(user_group_pd_path, default_user_group_pd_fln)
    else:
        user_group_pd_path = os.path.join(startup_dir, default_user_group_pd_fln)
    if not os.path.isfile(user_group_pd_path):
        logger.error(
            "The file with user permissions was not found at "
            "'%s'. User groups are not defined. USERS WILL NOT BE ABLE TO SUBMIT PLANS OR "
            "EXECUTE ANY OTHER OPERATIONS THAT REQUIRE PERMISSIONS.",
            user_group_pd_path,
        )
        user_group_pd_path = None

    config_worker["existing_plans_and_devices_path"] = existing_pd_path
    config_manager["existing_plans_and_devices_path"] = existing_pd_path
    config_manager["user_group_permissions_path"] = user_group_pd_path

    config_worker["update_existing_plans_devices"] = args.update_existing_plans_devices

    config_manager["user_group_permissions_reload"] = args.user_group_permissions_reload

    # Read private key from the environment variable, then check if the CLI parameter exists
    zmq_private_key = os.environ.get("QSERVER_ZMQ_PRIVATE_KEY", None)
    zmq_private_key = zmq_private_key if zmq_private_key else None  # Case of key==""
    if zmq_private_key is not None:
        try:
            validate_zmq_key(zmq_private_key)
        except Exception as ex:
            logger.error("ZMQ private key is improperly formatted: %s", str(ex))
            return 1

    config_manager["zmq_addr"] = args.zmq_addr
    config_manager["zmq_private_key"] = zmq_private_key

    redis_addr = args.redis_addr
    if redis_addr.count(":") > 1:
        logger.error(f"Redis address is incorrectly formatted: '{redis_addr}'")
        return 1
    config_manager["redis_addr"] = redis_addr

    wp = WatchdogProcess(
        config_worker=config_worker, config_manager=config_manager, msg_queue=msg_queue, log_level=log_level
    )
    try:
        wp.run()
    except KeyboardInterrupt:
        logger.info("The program was manually stopped")



import uuid


def format_jsonrpc_msg(method, params=None, *, notification=False):
    """
    Returns dictionary that contains JSON RPC message.

    Parameters
    ----------
    method: str
        Method name
    params: dict or list, optional
        List of args or dictionary of kwargs.
    notification: boolean
        If the message is notification, no response will be expected.
    """
    msg = {"method": method, "jsonrpc": "2.0"}
    if params is not None:
        msg["params"] = params
    if not notification:
        msg["id"] = str(uuid.uuid4())
    return msg



# flake8: noqa
import time as ttime
import typing
import ophyd
from ophyd import Device, Component as Cpt  # Keep 'Device' imported, used in unit tests
import bluesky
import bluesky.preprocessors as bpp
import bluesky.plan_stubs as bps
from bluesky_queueserver.manager.annotation_decorator import parameter_annotation_decorator


# Some useless devices for unit tests.
custom_test_device = ophyd.Device(name="custom_test_device")
custom_test_signal = ophyd.Signal(name="custom_test_signal")
custom_test_flyer = ophyd.sim.MockFlyer("custom_test_flyer", ophyd.sim.det, ophyd.sim.motor, 1, 5, 20)


@parameter_annotation_decorator(
    {
        "description": "Move motors into positions; then count dets.",
        "parameters": {
            "motors": {
                "description": "List of motors to be moved into specified positions before the measurement",
                "annotation": "typing.List[Motors]",
                "devices": {"Motors": ("motor1", "motor2")},
            },
            "detectors": {
                "description": "Detectors to use for measurement.",
                "annotation": "typing.List[Detectors]",
                "devices": {"Detectors": ("det1", "det2", "det3")},
                "default": ["det1", "det2"],
            },
            "positions": {
                "description": "Motor positions. The number of positions must be equal "
                "to the number of the motors.",
                "annotation": "typing.List[float]",
                "min": -10,
                "max": 10,
                "step": 0.01,
            },
        },
    }
)
def move_then_count(
    motors: typing.List[ophyd.device.Device],
    detectors: typing.Optional[typing.List[ophyd.device.Device]] = None,
    positions: typing.Optional[typing.List[float]] = None,
) -> typing.Generator[bluesky.utils.Msg, None, None]:
    if not isinstance(motors, (list, tuple)):
        raise TypeError(f"Parameter 'motors' should be a list or a tuple: type(motors) = {type(motors)}")
    if not isinstance(detectors, (list, tuple)):
        raise TypeError(f"Parameter 'detectors' should be a list or a tuple: type(detectors) = {type(detectors)}")
    if not isinstance(positions, (list, tuple)):
        raise TypeError(f"Parameter 'positions' should be a list or a tuple: type(positions) = {type(positions)}")
    if len(motors) != len(positions):
        raise TypeError(
            f"The lists of 'motors' and 'positions' should have the same number of elements: "
            f"len(motors) = {len(motors)}, len(positions) = {len(positions)}"
        )

    mv_args = [val for tup in zip(motors, positions) for val in tup]
    yield from bps.mv(*mv_args)
    yield from count(detectors)


@bpp.set_run_key_decorator("run_2")
@bpp.run_decorator(md={})
def _sim_plan_inner(npts: int, delay: float = 1.0):
    for j in range(npts):
        yield from bps.mov(motor1, j * 0.1 + 1, motor2, j * 0.2 - 2)
        yield from bps.trigger_and_read([motor1, motor2, det2])
        yield from bps.sleep(delay)


@bpp.set_run_key_decorator("run_1")
@bpp.run_decorator(md={})
@parameter_annotation_decorator(
    {
        "description": "Simulated multi-run plan: two nested runs. "
        "The plan is included for testing purposes only.",
        "parameters": {
            "npts": {
                "description": "The number of measurements in the outer run. "
                "Inner run will contain 'npts+1' measurements.",
            },
            "delay": {
                "description": "Delay between measurements.",
            },
        },
    }
)
def sim_multirun_plan_nested(npts: int, delay: float = 1.0):
    for j in range(int(npts / 2)):
        yield from bps.mov(motor, j * 0.2)
        yield from bps.trigger_and_read([motor, det])
        yield from bps.sleep(delay)

    yield from _sim_plan_inner(npts + 1, delay)

    for j in range(int(npts / 2), npts):
        yield from bps.mov(motor, j * 0.2)
        yield from bps.trigger_and_read([motor, det])
        yield from bps.sleep(delay)


# =====================================================================================
#                Functions for testing 'function_execute' API.
#
#        NOTE: those functions are used in unit tests. Changing the functions
#                     may cause those tests to fail.


def function_sleep(time):
    """
    Sleep for a given number of seconds.
    """
    print("******** Starting execution of the function 'function_sleep' **************")
    print(f"*******************   Waiting for {time} seconds **************************")
    ttime.sleep(time)
    print("******** Finished execution of the function 'function_sleep' **************")

    return {"success": True, "time": time}


_fifo_buffer = []


def push_buffer_element(element):
    """
    Push an element to a FIFO buffer.
    """
    print("******** Executing the function 'push_buffer_element' **************")
    _fifo_buffer.append(element)


def pop_buffer_element():
    """
    Pop an element from FIFO buffer. Raises exception if the buffer is empty
    (function call fails, traceback should be returned).
    """
    print("******** Executing the function 'pop_buffer_element' **************")
    return _fifo_buffer.pop(0)


def clear_buffer():
    """
    The function used for testing of 'function_execute' API.
    """
    print("******** Executing the function 'clear_buffer' **************")
    return _fifo_buffer.clear()


# ===========================================================================================
#     Simulated devices with subdevices. The devices are used in unit tests. Do not delete.
#     If class names are changed, search and change the names in unit tests.
#     Formatting of imported classes is inconsistent (e.g. 'Device', 'ophyd.Device' and
#     'ophyd.sim.SynAxis') is inconsistent on purpose to check if all possible versions work.


class SimStage(Device):
    x = Cpt(ophyd.sim.SynAxis, name="y", labels={"motors"})
    y = Cpt(ophyd.sim.SynAxis, name="y", labels={"motors"})
    z = Cpt(ophyd.sim.SynAxis, name="z", labels={"motors"})

    def set(self, x, y, z):
        """Makes the device Movable"""
        self.x.set(x)
        self.y.set(y)
        self.z.set(z)


class SimDetectors(Device):
    """
    The detectors are controlled by simulated 'motor1' and 'motor2'
    defined on the global scale.
    """

    det_A = Cpt(
        ophyd.sim.SynGauss,
        name="det_A",
        motor=motor1,
        motor_field="motor1",
        center=0,
        Imax=5,
        sigma=0.5,
        labels={"detectors"},
    )
    det_B = Cpt(
        ophyd.sim.SynGauss,
        name="det_B",
        motor=motor2,
        motor_field="motor2",
        center=0,
        Imax=5,
        sigma=0.5,
        labels={"detectors"},
    )


class SimBundle(ophyd.Device):
    mtrs = Cpt(SimStage, name="mtrs")
    dets = Cpt(SimDetectors, name="dets")


sim_bundle_A = SimBundle(name="sim_bundle_A")
sim_bundle_B = SimBundle(name="sim_bundle_B")  # Used for tests


@parameter_annotation_decorator(
    {
        "parameters": {
            "detectors": {  # Annotation for the parameter 'detectors'
                "annotation": "typing.List[DetList]",
                "devices": {"DetList": ["det", "det1", "det2", "det3", ":+^sim_bundle:?.*$"]},
            }
        }
    }
)
def count_bundle_test(detectors, num=1, delay=None, *, per_shot=None, md=None):
    yield from count(detectors, num=num, delay=delay, per_shot=per_shot, md=md)


# =======================================================================================
